#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#


import sys
import numpy
from pyscf.lib import logger
from pyscf import gto
from pyscf.gto.basis import _format_basis_name
from pyscf import ao2mo
from pyscf.data import elements
from pyscf.lib.exceptions import BasisNotFoundError
from pyscf.df.autoaux import autoaux, autoabs
from pyscf import __config__

DFBASIS = getattr(__config__, 'df_addons_aug_etb_beta', 'weigend')
ETB_BETA = getattr(__config__, 'df_addons_aug_dfbasis', 2.0)
FIRST_ETB_ELEMENT = getattr(__config__, 'df_addons_aug_start_at', 36)  # 'Rb'

# TODO: Switch to other default scheme for auxiliary basis generation.
# The auxiliary basis set generated by version 2.6 (and earlier) lacks compact
# functions. It may cause higher errors in ERI integrals.
USE_VERSION_26_AUXBASIS = True

# Obtained from http://www.psicode.org/psi4manual/master/basissets_byfamily.html
DEFAULT_AUXBASIS = {
    # AO basis       JK-fit                     MP2-fit
    'ccpvdz'      : ('cc-pvdz-jkfit'          , 'cc-pvdz-ri'         ),
    'augccpvdz'   : ('aug-cc-pvdz-jkfit'      , 'aug-cc-pvdz-ri'     ),
    'ccpvtz'      : ('cc-pvtz-jkfit'          , 'cc-pvtz-ri'         ),
    'augccpvtz'   : ('aug-cc-pvtz-jkfit'      , 'aug-cc-pvtz-ri'     ),
    'ccpvqz'      : ('cc-pvqz-jkfit'          , 'cc-pvqz-ri'         ),
    'augccpvqz'   : ('aug-cc-pvqz-jkfit'      , 'aug-cc-pvqz-ri'     ),
    'ccpv5z'      : ('cc-pv5z-jkfit'          , 'cc-pv5z-ri'         ),
    'augccpv5z'   : ('aug-cc-pv5z-jkfit'      , 'aug-cc-pv5z-ri'     ),
    'def2svp'     : ('def2-svp-jkfit'         , 'def2-svp-ri'        ),
    'def2svpd'    : ('def2-svp-jkfit'         , 'def2-svpd-ri'       ),
    'def2tzvp'    : ('def2-tzvp-jkfit'        , 'def2-tzvp-ri'       ),
    'def2mtzvp'   : ('def2-tzvp-jkfit'        , 'def2-tzvp-ri'       ),
    'def2tzvpd'   : ('def2-tzvp-jkfit'        , 'def2-tzvpd-ri'      ),
    'def2tzvpp'   : ('def2-tzvpp-jkfit'       , 'def2-tzvpp-ri'      ),
    'def2mtzvpp'  : ('def2-tzvpp-jkfit'       , 'def2-tzvpp-ri'      ),
    'def2tzvppd'  : ('def2-tzvpp-jkfit'       , 'def2-tzvppd-ri'     ),
    'def2qzvp'    : ('def2-qzvp-jkfit'        , 'def2-qzvp-ri'       ),
    #'def2qzvpd'   : ('def2-qzvp-jkfit'        , None                 ),
    'def2qzvpp'   : ('def2-qzvpp-jkfit'       , 'def2-qzvpp-ri'      ),
    'def2qzvppd'  : ('def2-qzvpp-jkfit'       , 'def2-qzvppd-ri'     ),
    'sto3g'       : ('def2-svp-jkfit'         , 'def2-svp-ri'        ),
    '321g'        : ('def2-svp-jkfit'         , 'def2-svp-ri'        ),
    '631g'        : ('cc-pvdz-jkfit'          , 'cc-pvdz-ri'         ),
    '631+g'       : ('heavy-aug-cc-pvdz-jkfit', 'heavyaug-cc-pvdz-ri'),
    '631++g'      : ('aug-cc-pvdz-jkfit'      , 'aug-cc-pvdz-ri'     ),
    '6311g'       : ('cc-pvtz-jkfit'          , 'cc-pvtz-ri'         ),
    '6311+g'      : ('heavy-aug-cc-pvtz-jkfit', 'heavyaug-cc-pvtz-ri'),
    '6311++g'     : ('aug-cc-pvtz-jkfit'      , 'aug-cc-pvtz-ri'     ),
}

class load(ao2mo.load):
    '''load 3c2e integrals from hdf5 file. It can be used in the context
    manager:

    with load(cderifile) as eri:
        print(eri.shape)
    '''
    def __init__(self, eri, dataname='j3c'):
        ao2mo.load.__init__(self, eri, dataname)

def _aug_etb_element(nuc_charge, basis, beta):
    l_max = max(b[0] for b in basis)
    emin_by_l = [1e99] * (l_max+1)
    emax_by_l = [0] * (l_max+1)
    for b in basis:
        l = b[0]
        if isinstance(b[1], (int, numpy.integer)):
            e_c = numpy.array(b[2:])
        else:
            e_c = numpy.array(b[1:])
        es = e_c[:,0]
        cs = e_c[:,1:]
        es = es[abs(cs).max(axis=1) > 1e-3]
        emax_by_l[l] = max(es.max(), emax_by_l[l])
        emin_by_l[l] = min(es.min(), emin_by_l[l])

    conf = elements.CONFIGURATION[nuc_charge]
    # 1: H - Be, 2: B - Ca, 3: Sc - La, 4: Ce -
    max_shells = 4 - conf.count(0)

    if USE_VERSION_26_AUXBASIS:
        # This is the method that version 2.6 (and earlier) generates auxiliary
        # basis. It estimates the exponents ranges by geometric average.
        # This method is not recommended because it tends to generate diffuse
        # functions. Important compact functions might be improperly excluded.
        l_max = min(l_max, max_shells)
        l_max_aux = l_max * 2
        l_max1 = l_max + 1
        emin_by_l = numpy.array(emin_by_l[:l_max1])
        emax_by_l = numpy.array(emax_by_l[:l_max1])
        emax = (emax_by_l[:,None] * emax_by_l) ** .5 * 2
        emin = (emin_by_l[:,None] * emin_by_l) ** .5 * 2
    else:
        # Using normal average, more auxiliary functions, especially compact
        # functions, will be generated.
        l_max_aux = min(l_max, max_shells) * 2
        l_max1 = l_max + 1
        emin_by_l = numpy.array(emin_by_l)
        emax_by_l = numpy.array(emax_by_l)
        emax = emax_by_l[:,None] + emax_by_l
        emin = emin_by_l[:,None] + emin_by_l

    liljsum = numpy.arange(l_max1)[:,None] + numpy.arange(l_max1)
    emax_by_l = numpy.array([emax[liljsum==ll].max() for ll in range(l_max_aux+1)])
    emin_by_l = numpy.array([emin[liljsum==ll].min() for ll in range(l_max_aux+1)])

    ns = numpy.log((emax_by_l+emin_by_l)/emin_by_l) / numpy.log(beta)
    etb = []
    for l, n in enumerate(numpy.ceil(ns).astype(int)):
        if n > 0:
            etb.append((l, n, emin_by_l[l], beta))
    return etb

def aug_etb_for_dfbasis(mol, dfbasis=DFBASIS, beta=ETB_BETA,
                        start_at=FIRST_ETB_ELEMENT):
    '''augment weigend basis with even-tempered gaussian basis
    exps = alpha*beta^i for i = 1..N
    '''
    nuc_start = gto.charge(start_at)
    uniq_atoms = {a[0] for a in mol._atom}

    newbasis = {}
    for symb in uniq_atoms:
        nuc_charge = gto.charge(symb)
        if nuc_charge < nuc_start:
            newbasis[symb] = dfbasis
        else:
            basis = mol._basis[symb]
            etb = _aug_etb_element(nuc_charge, basis, beta)
            if etb:
                newbasis[symb] = gto.expand_etbs(etb)
                for l, n, emin, beta in etb:
                    logger.info(mol, 'ETB for %s: l = %d, exps = %s * %g^n , n = 0..%d',
                                symb, l, emin, beta, n-1)
            else:
                raise RuntimeError(f'Failed to generate even-tempered auxbasis for {symb}')

    return newbasis

def aug_etb(mol, beta=ETB_BETA):
    '''To generate the even-tempered auxiliary Gaussian basis'''
    return aug_etb_for_dfbasis(mol, beta=beta, start_at=0)

def make_auxbasis(mol, *, xc='HF', mp2fit=False):
    '''Depending on the orbital basis, generating even-tempered Gaussians or
    the optimized auxiliary basis defined in DEFAULT_AUXBASIS
    '''
    uniq_atoms = {a[0] for a in mol._atom}
    auxbasis = {}
    if isinstance(mol.basis, str):
        _basis = {a: mol.basis for a in uniq_atoms}
        default_auxbasis = predefined_auxbasis(mol, mol.basis, xc, mp2fit)
        if default_auxbasis:
            auxbasis = {a: default_auxbasis for a in uniq_atoms}
            logger.debug(mol, 'Default auxbasis %s is applied universally', default_auxbasis)
    elif isinstance(mol.basis, dict):
        if 'default' in mol.basis:
            default_basis = mol.basis['default']
            _basis = {a: default_basis for a in uniq_atoms}
            _basis.update(mol.basis)
            del _basis['default']
        else:
            _basis = mol.basis
    else:
        _basis = mol._basis or {}

    for k, obs in _basis.items():
        if not isinstance(obs, str):
            continue
        if k in auxbasis:
            auxb = auxbasis[k]
            try:
                # Test if basis auxb for element k is available
                gto.basis.load(auxb, elements._std_symbol_without_ghost(k))
            except BasisNotFoundError:
                del auxbasis[k]
        else:
            balias = _format_basis_name(obs)
            if gto.basis._is_pople_basis(balias):
                balias = balias.split('g')[0] + 'g'
            auxb = predefined_auxbasis(mol, balias, xc, mp2fit)
            if auxb is not None:
                try:
                    gto.basis.load(auxb, elements._std_symbol_without_ghost(k))
                except BasisNotFoundError:
                    continue
                auxbasis[k] = auxb
                logger.info(mol, 'Default auxbasis %s is used for %s %s',
                            auxb, k, obs)

    if len(auxbasis) != len(_basis):
        # Some AO basis not found in DEFAULT_AUXBASIS
        auxbasis, auxdefault = aug_etb(mol), auxbasis
        auxbasis.update(auxdefault)
        aux_etb = set(auxbasis) - set(auxdefault)
        if aux_etb:
            logger.warn(mol, 'Even tempered Gaussians are generated as '
                        'DF auxbasis for  %s', ' '.join(aux_etb))
            for k in aux_etb:
                logger.debug(mol, '  ETB auxbasis for %s  %s', k, auxbasis[k])
    return auxbasis

# TODO: add auxbasis keyword etb and auto
def make_auxmol(mol, auxbasis=None):
    '''Generate a fake Mole object which uses the density fitting auxbasis as
    the basis sets.  If auxbasis is not specified, the optimized auxiliary fitting
    basis set will be generated according to the rules recorded in
    pyscf.df.addons.DEFAULT_AUXBASIS.  If the optimized auxiliary basis is not
    available (either not specified in DEFAULT_AUXBASIS or the basis set of the
    required elements not defined in the optimized auxiliary basis),
    even-tempered Gaussian basis set will be generated.

    See also the paper JCTC, 13, 554 about generating auxiliary fitting basis.

    Kwargs:
        auxbasis : str, list, tuple
            Similar to the input of orbital basis in Mole object.
    '''
    pmol = mol.copy(deep=False)

    if auxbasis is None:
        auxbasis = make_auxbasis(mol)
    pmol.basis = auxbasis

    if isinstance(auxbasis, (str, list, tuple)):
        uniq_atoms = {a[0] for a in mol._atom}
        _basis = {a: auxbasis for a in uniq_atoms}
    else:
        assert isinstance(auxbasis, dict)
        if 'default' in auxbasis:
            uniq_atoms = {a[0] for a in mol._atom}
            _basis = {a: auxbasis['default'] for a in uniq_atoms}
            _basis.update(auxbasis)
            del (_basis['default'])
        else:
            _basis = auxbasis

    try:
        pmol._basis = pmol.format_basis(_basis)
    except BasisNotFoundError:
        if isinstance(auxbasis, str):
            print(f'''
Some elements are not found in the specified auxiliary basis set {auxbasis}.
To proceed, you can generate even-tempered Gaussian (ETB) functions for the missing elements.
Three routines are available for this system:
1. PySCF's built-in basis generation function
    mf.with_df.auxbasis = pyscf.df.make_auxbasis(mol)
2. ORCA recommended AutoAux ETB
    mf.with_df.auxbasis = pyscf.df.autoaux(mol)
3. Gaussian package recommended ETB
    mf.with_df.auxbasis = pyscf.df.autoabs(mol)
''')
        raise

    # Note: To pass parameters like gauge origin, rsh-omega to auxmol,
    # mol._env[:PTR_ENV_START] must be copied to auxmol._env
    pmol._atm, pmol._bas, pmol._env = \
            pmol.make_env(mol._atom, pmol._basis, mol._env[:gto.PTR_ENV_START])
    pmol._built = True
    logger.debug(mol, 'num shells = %d, num cGTOs = %d',
                 pmol.nbas, pmol.nao_nr())
    return pmol

def bse_predefined_auxbasis(mol, basis, xc='HF', mp2fit=False):
    '''Find auxiliary basis sets for XC functionals from BSE database.
    If no matching basis set is found, the function returns None.
    '''
    if not isinstance(basis, str):
        return None

    try:
        from pyscf.dft.libxc import is_hybrid_xc
    except ImportError:
        from pyscf.dft.xcfun import is_hybrid_xc
    pyscf_basis_alias = _format_basis_name(basis).lower()
    basis_meta = gto.mole.BSE_META.get(pyscf_basis_alias)
    auxbasis = None
    if basis_meta:
        auxiliaries = basis_meta[2]
        if mp2fit:
            auxbasis = auxiliaries.get('rifit')
            if auxbasis:
                logger.debug(mol, f'BSE predefined RIFIT basis set {auxbasis} for mp2fit')
        elif is_hybrid_xc(xc):
            auxbasis = auxiliaries.get('jkfit')
            if auxbasis:
                logger.debug(mol, f'BSE predefined JKFIT basis set {auxbasis} for {xc}')
        else:
            auxbasis = auxiliaries.get('jfit')
            if auxbasis is None:
                auxbasis = auxiliaries.get('dftjfit')
            if auxbasis is None:
                auxbasis = auxiliaries.get('jkfit')
            if auxbasis:
                logger.debug(mol, f'BSE predefined JFIT basis set {auxbasis} for {xc}')
    return auxbasis

def predefined_auxbasis(mol, basis, xc='HF', mp2fit=False):
    '''Predefined auxiliary basis sets for the specified orbital basis set and
    XC functional. The searching starts from the Psi4 recommendation. If not
    found, the record in BSE database will be used. If no matching basis set is
    found, the function returns None.
    '''
    if not isinstance(basis, str):
        return None

    try:
        from pyscf.dft.libxc import is_hybrid_xc
    except ImportError:
        from pyscf.dft.xcfun import is_hybrid_xc
    pyscf_basis_alias = _format_basis_name(basis).lower()
    if pyscf_basis_alias in DEFAULT_AUXBASIS:
        if mp2fit:
            auxbasis = DEFAULT_AUXBASIS[pyscf_basis_alias][1]
            logger.debug(mol, f'Psi4 predefined RIFIT basis set {auxbasis} for mp2fit')
            return auxbasis
        elif is_hybrid_xc(xc):
            auxbasis = DEFAULT_AUXBASIS[pyscf_basis_alias][0]
            logger.debug(mol, f'Psi4 predefined JKFIT basis set {auxbasis} for {xc}')
            return auxbasis
    return bse_predefined_auxbasis(mol, basis, xc, mp2fit)

del (DFBASIS, ETB_BETA, FIRST_ETB_ELEMENT)



# Copyright 2023 The GPU4PySCF Authors. All Rights Reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Gradient of SMD solvent model (for experiment and education)
copied from GPU4PySCF with modification for CPU
'''

import numpy as np
from pyscf.data import radii
from pyscf.solvent import pcm, smd
from pyscf.solvent.grad import pcm as pcm_grad
from pyscf.solvent.smd import (
    sigma_water, sigma_n, sigma_alpha, sigma_beta, r_zz, swtich_function,
    hartree2kcal)

def grad_swtich_function(R, r, dr):
    if R < r + dr:
        return -np.exp(dr/(R-dr-r)) * dr / (R-dr-r)**2
    else:
        return 0.0

def atomic_surface_tension(symbols, coords, n, alpha, beta, water=True):
    '''
    - list of atomic symbols
    - atomic coordinates in Anstrong
    - solvent descriptors: n, alpha, beta
    '''

    def get_bond_tension(bond):
        if water:
            return sigma_water.get(bond, 0.0)
        t = 0.0
        t += sigma_n.get(bond, 0.0) * n
        t += sigma_alpha.get(bond, 0.0) * alpha
        t += sigma_beta.get(bond, 0.0) * beta
        return t

    def get_atom_tension(sym_i):
        if water:
            return sigma_water.get(sym_i, 0.0)
        t = 0.0
        t += sigma_n.get(sym_i, 0.0) * n
        t += sigma_alpha.get(sym_i, 0.0) * alpha
        t += sigma_beta.get(sym_i, 0.0) * beta
        return t
    natm = coords.shape[0]
    ri_rj = coords[:,None,:] - coords[None,:,:]
    rij = np.sum(ri_rj**2, axis=2)**0.5
    np.fill_diagonal(rij, 1)
    drij = ri_rj / np.expand_dims(rij, axis=-1)
    tensions = []
    for i, sym_i in enumerate(symbols):
        if sym_i not in ['H', 'C', 'N', 'O', 'F', 'Si', 'S', 'Cl', 'Br']:
            tensions.append(np.zeros([natm,3]))
            continue

        tension = np.zeros([natm,3])
        if sym_i in ['F', 'Si', 'S', 'Cl', 'Br']:
            tensions.append(tension)
            continue

        if sym_i == 'H':
            dt_HC = np.zeros([natm,3])
            dt_HO = np.zeros([natm,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('H','C'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_HC[i] += dt_drij
                    dt_HC[j] -= dt_drij
                if sym_j == 'O':
                    r, dr = r_zz.get(('H','O'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_HO[i] += dt_drij
                    dt_HO[j] -= dt_drij
            sig_HC = get_bond_tension(('H','C'))
            sig_HO = get_bond_tension(('H','O'))
            tension += sig_HC * dt_HC + sig_HO * dt_HO
            tensions.append(tension)
            continue

        if sym_i == 'C':
            dt_CC = np.zeros([natm,3])
            dt_CN = np.zeros([natm,3])
            t_CN = 0.0
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C' and i != j:
                    r, dr = r_zz.get(('C', 'C'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_CC[i] += dt_drij
                    dt_CC[j] -= dt_drij
                if sym_j == 'N':
                    r, dr = r_zz.get(('C', 'N'), (0.0, 0.0))
                    t_CN += swtich_function(rij[i,j], r, dr)
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_CN[i] += dt_drij
                    dt_CN[j] -= dt_drij
            sig_CC = get_bond_tension(('C','C'))
            sig_CN = get_bond_tension(('C','N'))
            tension += sig_CC * dt_CC + sig_CN * (2 * t_CN * dt_CN)
            tensions.append(tension)
            continue

        if sym_i == 'N':
            t_NC = 0.0
            dt_NC = np.zeros([natm,3])
            dt_NC3 = np.zeros([natm,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('N','C'), (0.0, 0.0))
                    tk = 0.0
                    dtk = np.zeros([natm,3])
                    for k, sym_k in enumerate(symbols):
                        if k != i and k != j:
                            rjk, drjk = r_zz.get(('C', sym_k), (0.0, 0.0))
                            tk += swtich_function(rij[j,k], rjk, drjk)
                            dtk_rjk = grad_swtich_function(rij[j,k], rjk, drjk) * drij[j,k]
                            dtk[j] += dtk_rjk
                            dtk[k] -= dtk_rjk

                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j] * tk**2
                    dt_NC[i] += dt_drij
                    dt_NC[j] -= dt_drij

                    t = swtich_function(rij[i,j], r, dr)
                    dt_NC += t * (2 * tk * dtk)
                    t_NC += t * tk**2

                    r, dr = r_zz.get(('N','C3'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_NC3[i] += dt_drij
                    dt_NC3[j] -= dt_drij
            sig_NC = get_bond_tension(('N','C'))
            sig_NC3= get_bond_tension(('N','C3'))
            tension += sig_NC * (1.3 * t_NC**0.3 * dt_NC) + sig_NC3 * dt_NC3
            tensions.append(tension)
            continue

        if sym_i == 'O':
            dt_OC = np.zeros([natm,3])
            dt_ON = np.zeros([natm,3])
            dt_OO = np.zeros([natm,3])
            dt_OP = np.zeros([natm,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('O','C'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_OC[i] += dt_drij
                    dt_OC[j] -= dt_drij
                if sym_j == 'N':
                    r, dr = r_zz.get(('O','N'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_ON[i] += dt_drij
                    dt_ON[j] -= dt_drij
                if sym_j == 'O' and j != i:
                    r, dr = r_zz.get(('O','O'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_OO[i] += dt_drij
                    dt_OO[j] -= dt_drij
                if sym_j == 'P':
                    r, dr = r_zz.get(('O','P'), (0.0, 0.0))
                    dt_drij = grad_swtich_function(rij[i,j], r, dr) * drij[i,j]
                    dt_OP[i] += dt_drij
                    dt_OP[j] -= dt_drij
            sig_OC = get_bond_tension(('O','C'))
            sig_ON = get_bond_tension(('O','N'))
            sig_OO = get_bond_tension(('O','O'))
            sig_OP = get_bond_tension(('O','P'))
            tension += sig_OC * dt_OC + sig_ON * dt_ON + sig_OO * dt_OO + sig_OP * dt_OP
            tensions.append(tension)
            continue

    return np.asarray(tensions)

def get_cds(smdobj):
    mol = smdobj.mol
    n, _, alpha, beta, gamma, _, phi, psi = smdobj.solvent_descriptors
    symbols = [mol.atom_symbol(ia) for ia in range(mol.natm)]
    coords = mol.atom_coords(unit='A')
    if smdobj._solvent.lower() != 'water':
        grad_tension = atomic_surface_tension(symbols, coords, n, alpha, beta, water=False)
        atm_tension = smd.atomic_surface_tension(symbols, coords, n, alpha, beta, water=False)
        mol_tension = smd.molecular_surface_tension(beta, gamma, phi, psi)
    else:
        grad_tension = atomic_surface_tension(symbols, coords, n, alpha, beta, water=True)
        atm_tension = smd.atomic_surface_tension(symbols, coords, n, alpha, beta, water=True)
        mol_tension = 0.0

    # generate surface for calculating SASA
    rad = radii.VDW + 0.4/radii.BOHR
    surface = pcm.gen_surface(mol, ng=smdobj.sasa_ng, rad=rad)
    _, grad_area = pcm_grad.get_dF_dA(surface)
    area = surface['area']
    gridslice = surface['gslice_by_atom']
    SASA = np.asarray([np.sum(area[p0:p1], axis=0) for p0,p1, in gridslice])
    grad_SASA = np.asarray([np.sum(grad_area[p0:p1], axis=0) for p0,p1, in gridslice])
    SASA *= radii.BOHR**2
    grad_SASA *= radii.BOHR**2
    mol_cds = mol_tension * np.sum(grad_SASA, axis=0) / 1000
    grad_tension *= radii.BOHR
    atm_cds = np.einsum('i,ijx->jx', SASA, grad_tension) / 1000
    atm_cds+= np.einsum('i,ijx->jx', atm_tension, grad_SASA) / 1000
    return (mol_cds + atm_cds)/hartree2kcal # hartree



#!/usr/bin/env python
# Copyright 2020-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.gto import mole
from functools import reduce

SYMPREC = getattr(__config__, 'pbc_symm_space_group_symprec', 1e-6) #this has the unit of length

def search_point_group_ops(cell, tol=SYMPREC):
    a = cell.lattice_vectors()
    G = np.dot(a, a.T)
    pbc_axis = np.array([1,1,1], dtype=bool)
    if cell.dimension < 3:
        pbc_axis[cell.dimension:] = False

    a_norm = np.sqrt(np.diag(G))
    a_angle = np.arccos(G / np.outer(a_norm, a_norm))
    tol2 = tol**2

    rotations = []
    for op in lib.cartesian_prod([[1,0,-1],]*9):
        W = np.asarray(op, dtype=np.int32).reshape(3,3)
        G_tilde = reduce(np.dot, (W.T, G, W))

        #check change of metric
        a_tilde_norm = np.sqrt(np.diag(G_tilde))
        length_error = np.abs(a_norm - a_tilde_norm)
        if (length_error > tol).any():
            continue
        tmp = (a_norm + a_tilde_norm)
        a_tilde_angle = np.arccos(G_tilde / np.outer(a_tilde_norm, a_tilde_norm))
        angle_error = np.sin(a_angle - a_tilde_angle) **2 * np.outer(tmp,tmp) / 4
        if (angle_error > tol2).any():
            continue

        #check if rotation inverts non-periodic axes
        if not (W[np.diag(~pbc_axis)] == 1).all():
            continue

        #check if rotation swaps periodic and non-periodic axes
        pbc_axis2 = np.logical_and.outer(pbc_axis, pbc_axis)
        if W[~(pbc_axis2 | np.eye(3, dtype=bool))].any():
            continue

        rotations.append(W)

    rotations = np.asarray(rotations)
    return rotations

def search_space_group_ops(cell, rotations=None, tol=SYMPREC):
    '''
    Search for the allowed space group operations for a specific cell.

    Notes:
        The current implementation treats the cell with the spins on all
        atoms flipped as the same as the original cell. If this is not desired,
        then one can use different names for the two sets of atoms and set their
        magnetic moment to 0.
    '''
    if rotations is None: rotations = search_point_group_ops(cell, tol=tol)
    a = cell.lattice_vectors()
    coords = cell.get_scaled_atom_coords()
    atmgrp = mole.atom_types(cell._atom, magmom=cell.magmom)
    atmgrp_spin_inv = {} #spin up and down inverted
    has_spin = False
    for atm in atmgrp.keys():
        if atm[-2:] == '_u':
            has_spin = True
            atmgrp_spin_inv[atm] = atmgrp[atm[:-2]+'_d']
        elif atm[-2:] == '_d':
            has_spin = True
            atmgrp_spin_inv[atm] = atmgrp[atm[:-2]+'_u']
        else:
            atmgrp_spin_inv[atm] = atmgrp[atm]

    def test_trans(R, t, spin_inverse=False):
        for atm, idx in atmgrp.items():
            x = coords[idx]
            xt = np.dot(x, R.T) + t
            if not spin_inverse:
                x_xt = np.concatenate((x,xt))
            else:
                x_xt = np.concatenate((coords[atmgrp_spin_inv[atm]],xt))
            x_xt = np.mod(x_xt, 1)
            x_xt = np.round(x_xt, -np.log10(tol).astype(int))
            x_xt = np.mod(x_xt, 1)
            sorted_idx = np.lexsort(x_xt.T)
            x_xt = x_xt[sorted_idx]
            diff = np.dot(x_xt[::2] - x_xt[1::2], a)
            if (np.abs(diff) > tol).any():
                return False
        return True

    grp_len = [len(v) for v in atmgrp.values()]
    atm = [k for k in atmgrp.keys() if len(atmgrp[k]) == min(grp_len)][0]
    x = coords[atmgrp[atm]]
    x_spin_inv = None
    if atm[-2:] in ['_u', '_d']:
        x_spin_inv = coords[atmgrp_spin_inv[atm]]

    from pyscf.pbc.symm.space_group import SPGElement
    ops = []
    for rot in rotations:
        w = x - np.dot(x[0], rot.T)
        if x_spin_inv is not None:
            w_spin_inv = x_spin_inv - np.dot(x[0], rot.T)
            w = np.vstack((w, w_spin_inv))
        w = np.mod(w, 1)
        w = np.round(w, -np.log10(tol).astype(int))
        w = np.mod(w, 1)
        w = np.unique(w, axis=0)
        for trans in w:
            if test_trans(rot, trans):
                ops.append(SPGElement(rot, trans))
            elif has_spin:
                if test_trans(rot, trans, True):
                    ops.append(SPGElement(rot, trans))
    return ops

def get_crystal_class(cell, ops=None, tol=SYMPREC):
    if ops is None: ops = search_space_group_ops(cell, tol=tol)
    rotations = []
    for op in ops:
        rotations.append(op.rot)
    rotations = np.unique(np.asarray(rotations), axis=0)

    maps =  {-6 : 0,
             -4 : 1,
             -3 : 2,
             -2 : 3,
             -1 : 4,
              1 : 5,
              2 : 6,
              3 : 7,
              4 : 8,
              6 : 9}
    table = [0,] * 10
    for rot in rotations:
        trace = np.trace(rot)
        det = np.linalg.det(rot)
        if trace == 3:
            assert(det == 1)
            table[maps[1]] += 1
        elif trace == -3:
            assert(det == -1)
            table[maps[-1]] += 1
        elif trace == 2:
            assert(det == 1)
            table[maps[6]] += 1
        elif trace == -2:
            assert(det == -1)
            table[maps[-6]] += 1
        elif trace == 0:
            if det == 1:
                table[maps[3]] += 1
            elif det == -1:
                table[maps[-3]] += 1
        elif trace == 1:
            if det == 1:
                table[maps[4]] += 1
            elif det == -1:
                table[maps[-2]] += 1
        elif trace == -1:
            if det == 1:
                table[maps[2]] += 1
            elif det == -1:
                table[maps[-4]] += 1
        else:
            raise ValueError("Input rotation matrix is wrong: %s" % rot)

    from pyscf.pbc.symm.tables import CrystalClass, LaueClass
    laue_class = None
    crystal_class = None
    for k, v in CrystalClass.items():
        count = 0
        for i in range(10):
            if table[i] == v[i]: count += 1
        if count == 10:
            crystal_class = k
            break
    for k, v in LaueClass.items():
        if crystal_class in v:
            laue_class = k
            break
    if crystal_class is None or laue_class is None:
        raise RuntimeError("Unable to determine crystal class.")
    return crystal_class, laue_class


if __name__ == "__main__":
    from pyscf.pbc import gto
    cell = gto.Cell()
    cell.atom = """
      Cu 1.000000     1.00000      0.0000
      O  0.000000     1.00000      0.0000
      O  1.000000     2.00000      0.0000
      Cu 1.000000     3.00000      0.0000
      O  1.000000     4.00000      0.0000
      O  2.000000     3.00000      0.0000
      Cu 3.000000     3.00000      0.0000
      O  4.000000     3.00000      0.0000
      O  3.000000     2.00000      0.0000
      Cu 3.000000     1.00000      0.0000
      O  3.000000     0.00000      0.0000
      O  2.000000     1.00000      0.0000
    """
    cell.a = [[4.0, 0., 0.], [0., 4.0, 0.], [0., 0., 16.0]]
    cell.dimension = 2
    cell.magmom = [1, 1, -1, -1, 1, -1, 1, 1, -1, -1, 1, -1]
    cell.build()

    ops = search_space_group_ops(cell)
    for op in ops: print(op)

    point_group = get_crystal_class(cell, ops=ops)[0]
    print(point_group)



# Copyright 2023 The GPU4PySCF Authors. All Rights Reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Hessian of SMD solvent model, copied from GPU4PySCF with modification for CPU
'''
# pylint: disable=C0103

import numpy as np
from pyscf import scf, lib
from pyscf.solvent import smd
from pyscf.solvent.grad import smd as smd_grad
from pyscf.solvent.grad import pcm as pcm_grad
from pyscf.solvent.hessian import pcm as pcm_hess
from pyscf.lib import logger

def get_cds(smdobj):
    mol = smdobj.mol
    solvent = smdobj.solvent
    def smd_grad_scanner(mol):
        smdobj_tmp = smd.SMD(mol)
        smdobj_tmp.solvent = solvent
        return smd_grad.get_cds(smdobj_tmp)

    log = logger.new_logger(mol, mol.verbose)
    t1 = (logger.process_clock(), logger.perf_counter())
    log.warn("Using finite difference scheme for CDS contribution.")
    eps = 1e-4
    natm = mol.natm
    hess_cds = np.zeros([natm,natm,3,3])
    for ia in range(mol.natm):
        for j in range(3):
            coords = mol.atom_coords(unit='B')
            coords[ia,j] += eps
            mol.set_geom_(coords, unit='B')
            mol.build()
            grad0_cds = smd_grad_scanner(mol)

            coords[ia,j] -= 2.0*eps
            mol.set_geom_(coords, unit='B')
            mol.build()
            grad1_cds = smd_grad_scanner(mol)

            coords[ia,j] += eps
            mol.set_geom_(coords, unit='B')
            hess_cds[ia,:,j] = (grad0_cds - grad1_cds) / (2.0 * eps)
    t1 = log.timer_debug1('solvent energy', *t1)
    return hess_cds # hartree



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import math
import numpy as np
from scipy import linalg
from pyscf import mcpdft
from pyscf.grad import mcpdft as mcpdft_grad
from pyscf import lib
from pyscf.lib import logger
from pyscf.fci import direct_spin1
from pyscf.mcscf import mc1step, newton_casscf
from pyscf.grad import rhf as rhf_grad
from pyscf.grad import casscf as casscf_grad
from pyscf.grad import sacasscf as sacasscf_grad
from pyscf import __config__
from itertools import product

# PySCF-Forge installation check
try:
    from pyscf.csf_fci.csf import CSFFCISolver
except ModuleNotFoundError:
    class CSFFCISolver:
        pass

CONV_TOL_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_conv_tol_diabatize', 1e-8)
SING_TOL_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_sing_tol_diabatize', 1e-8)
SING_STEP_TOL = getattr(__config__, 'grad_mspdft_sing_step_tol', 2*math.pi)

def _unpack_state (state):
    if hasattr (state, '__len__'): return state[0], state[1]
    return state, state

# TODO: state-average-mix generalization ?
def make_rdm12_heff_offdiag (mc, ci, si_bra, si_ket):
    '''Compute <bra|O|ket> - sum_i <i|O|i>, where O is the 1- and 2-RDM
    operator product, and |bra> and |ket> are both states spanning the
    vector space of |i>, which are multi-determinantal many-electron
    states in an active space.

    Args:
        mc : object of class CASCI or CASSCF
            Only "ncas" and "nelecas" are used, to determine Hilbert
            of ci
        ci : ndarray or list of length (nroots)
            Contains CI vectors spanning a model space
        si_bra : ndarray of shape (nroots)
            Coefficients of ci elements for state |bra>
        si_ket : ndarray of shape (nroots)
            Coefficients of ci elements for state |ket>

    Returns:
        casdm1 : ndarray of shape [ncas,]*2
            Contains O = p'q case
        casdm2 : ndarray of shape [ncas,]*4
            Contains O = p'q'sr case
    '''
    ncas, nelecas = mc.ncas, mc.nelecas
    nroots = len (ci)
    ci_arr = np.asarray (ci)
    ci_bra = np.tensordot (si_bra, ci_arr, axes=1)
    ci_ket = np.tensordot (si_ket, ci_arr, axes=1)
    casdm1, casdm2 = direct_spin1.trans_rdm12 (ci_bra, ci_ket, ncas, nelecas)
    ddm1 = np.zeros ((nroots, ncas, ncas), dtype=casdm1.dtype)
    ddm2 = np.zeros ((nroots, ncas, ncas, ncas, ncas), dtype=casdm1.dtype)
    for i in range (nroots):
        ddm1[i,...], ddm2[i,...] = direct_spin1.make_rdm12 (ci[i], ncas,
            nelecas)
    si_diag = si_bra * si_ket
    casdm1 -= np.tensordot (si_diag, ddm1, axes=1)
    casdm2 -= np.tensordot (si_diag, ddm2, axes=1)
    return casdm1, casdm2

# TODO: docstring?
def mspdft_heff_response (mc_grad, mo=None, ci=None,
        si_bra=None, si_ket=None, state=None,
        heff_mcscf=None, eris=None):
    '''Compute the orbital and intermediate-state rotation response
    vector in the context of an MS-PDFT gradient calculation '''
    mc = mc_grad.base
    if mo is None: mo = mc_grad.mo_coeff
    if ci is None: ci = mc_grad.ci
    if state is None: state = mc_grad.state
    ket, bra = _unpack_state (state)
    if si_bra is None: si_bra = mc.si[:,bra]
    if si_ket is None: si_ket = mc.si[:,ket]
    if heff_mcscf is None: heff_mcscf = mc.heff_mcscf
    if eris is None: eris = mc.ao2mo (mo)
    nroots, ncore = mc_grad.nroots, mc.ncore
    moH = mo.conj ().T

    # Orbital rotation (no all-core DM terms allowed!)
    # (Factor of 2 is convention difference between mc1step and newton_casscf)
    casdm1, casdm2 = make_rdm12_heff_offdiag (mc, ci, si_bra, si_ket)
    casdm1 = 0.5 * (casdm1 + casdm1.T)
    casdm2 = 0.5 * (casdm2 + casdm2.transpose (1,0,3,2))
    vnocore = eris.vhf_c.copy ()
    vnocore[:,:ncore] = -moH @ mc.get_hcore () @ mo[:,:ncore]
    with lib.temporary_env (eris, vhf_c=vnocore):
        g_orb = 2 * mc1step.gen_g_hop (mc, mo, 1, casdm1, casdm2, eris)[0]
    g_orb = mc.unpack_uniq_var (g_orb)

    # Intermediate state rotation (TODO: state-average-mix generalization)
    braH = np.dot (si_bra, heff_mcscf)
    Hket = np.dot (heff_mcscf, si_ket)
    si2 = si_bra * si_ket
    g_is  = np.multiply.outer (si_ket, braH)
    g_is += np.multiply.outer (si_bra, Hket)
    g_is -= 2 * si2[:,None] * heff_mcscf
    g_is -= g_is.T
    g_is = g_is[np.tril_indices (nroots, k=-1)]

    return g_orb, g_is

# TODO: docstring?
def mspdft_heff_HellmanFeynman (mc_grad, atmlst=None, mo=None, ci=None,
        si=None, si_bra=None, si_ket=None, state=None, eris=None, mf_grad=None,
        verbose=None, **kwargs):
    mc = mc_grad.base
    if atmlst is None: atmlst = mc_grad.atmlst
    if mo is None: mo = mc.mo_coeff
    if ci is None: ci = mc.ci
    if si is None: si = getattr (mc, 'si', None)
    if state is None: state = mc_grad.state
    ket, bra = _unpack_state (state)
    if si_bra is None: si_bra = si[:,bra]
    if si_ket is None: si_ket = si[:,ket]
    if eris is None: eris = mc.ao2mo (mo)
    if mf_grad is None: mf_grad = mc.get_rhf_base ().nuc_grad_method ()
    if verbose is None: verbose = mc_grad.verbose
    ncore = mc.ncore
    log = logger.new_logger (mc_grad, verbose)
    ci0 = np.zeros_like (ci[0])

    # CASSCF grad with effective RDMs
    t0 = (logger.process_clock (), logger.perf_counter ())
    casdm1, casdm2 = make_rdm12_heff_offdiag (mc, ci, si_bra, si_ket)
    casdm1 = 0.5 * (casdm1 + casdm1.T)
    casdm2 = 0.5 * (casdm2 + casdm2.transpose (1,0,3,2))
    dm12 = lambda * args: (casdm1, casdm2)
    fcasscf = mc_grad.make_fcasscf (state=ket,
        fcisolver_attr={'make_rdm12' : dm12})
    # TODO: DFeri functionality
    # Perhaps by patching fcasscf.nuc_grad_method?
    fcasscf_grad = fcasscf.nuc_grad_method ()
    #fcasscf_grad = casscf_grad.Gradients (fcasscf)
    de = fcasscf_grad.kernel (mo_coeff=mo, ci=ci0, atmlst=atmlst, verbose=0)

    # subtract nuc-nuc and core-core (patching out simplified gfock terms)
    moH = mo.conj ().T
    f0 = (moH @ mc.get_hcore () @ mo) + eris.vhf_c
    mo_energy = f0.diagonal ().copy ()
    mo_occ = np.zeros_like (mo_energy)
    mo_occ[:ncore] = 2.0
    f0 *= mo_occ[None,:]
    dme0 = lambda * args: mo @ ((f0+f0.T)*.5) @ moH
    with lib.temporary_env (mf_grad, make_rdm1e=dme0, verbose=0):
        with lib.temporary_env (mf_grad.base, mo_coeff=mo, mo_occ=mo_occ):
            # Second level there should become unnecessary in future, if anyone
            # ever gets around to cleaning up pyscf.df.grad.rhf & pyscf.grad.rhf
            dde = mf_grad.kernel (mo_coeff=mo, mo_energy=mo_energy, mo_occ=mo_occ,
                atmlst=atmlst)
    de -= dde
    log.debug ('MS-PDFT gradient off-diagonal H-F terms:\n{}'.format (de))
    log.timer ('MS-PDFT gradient off-diagonal H-F terms', *t0)
    return de

def get_diabfns (obj):
    '''Interpret the name of the MS-PDFT method as a pair of functions
    which compute the derivatives of a particular objective function
    with respect to wave function parameters and geometry perturbations,
    excluding first and second derivatives wrt intermediate state
    rotations, which is handled by the energy-class version of this
    function.

    Args:
        obj : string
            Specify particular MS-PDFT method. Currently, only "CMS" is
            supported. Not case-sensitive.

    Returns:
        diab_response : callable
            Computes the orbital-rotation and CI-transfer sectors of the
            Hessian-vector product of the MS objective function for a
            vector of intermediate-state rotations
        diab_grad : callable
            Computes the gradient of the MS objective function wrt
            geometry perturbation
    '''
    if obj.upper () == 'CMS':
        from pyscf.grad.cmspdft import diab_response, diab_grad
    else:
        raise RuntimeError ('MS-PDFT type not supported')
    return diab_response, diab_grad

# TODO: docstring? especially considering the "si_bra," "si_ket"
# functionality??
# TODO: figure out how to log the gradients with the right method name!
class Gradients (mcpdft_grad.Gradients):

    # Preconditioner solves the IS problem; hence, get_init_guess rewrite is
    # unnecessary
    get_init_guess = sacasscf_grad.Gradients.get_init_guess
    project_Aop = sacasscf_grad.Gradients.project_Aop

    def __init__(self, mc):
        self.conv_rtol = 0
        def_tol0 = getattr (mc, 'conv_tol_grad', None)
        if def_tol0 is None:
            def_tol0 = np.sqrt (getattr (mc, 'conv_tol', 1e-7))
        def_tol1 = getattr (mc, 'conv_tol_diabatize',
                            CONV_TOL_DIABATIZE)
        self.conv_atol = min (def_tol0, def_tol1)
        self.sing_step_tol = SING_STEP_TOL
        mcpdft_grad.Gradients.__init__(self, mc)
        r, g = get_diabfns (self.base.diabatization)
        self._diab_response = r
        self._diab_grad = g
        self.nlag += self.nis

    @property
    def nis (self):
        return self.nroots * (self.nroots - 1) // 2

    def diab_response (self, Lis, **kwargs):
        return self._diab_response (self, Lis, **kwargs)
    def diab_grad (self, Lis, **kwargs):
        return self._diab_grad (self, Lis, **kwargs)

    def kernel (self, state=None, mo=None, ci=None, si=None, _freeze_is=False,
            **kwargs):
        '''Cache the Hamiltonian and effective Hamiltonian terms, and
        pass around the IS hessian

        eris, veff1, veff2, and d2f should be available to all top-level
        functions: get_wfn_response, get_Aop_Adiag, get_ham_response,
        and get_LdotJnuc

        freeze_is == True sets the is component of the response to zero
        for debugging purposes
        '''
        if state is None: state = self.state
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if si is None: si = self.base.si
        if isinstance (ci, np.ndarray): ci = [ci] # hack hack hack...
        if state is None:
            raise NotImplementedError ('Gradient of PDFT state-average energy')
        self.state = state # Not the best code hygiene maybe
        nroots = self.nroots
        veff1 = []
        veff2 = []
        d2f = self.base.diabatizer (ci=ci)[2]
        for ix in range (nroots):
            v1, v2 = self.base.get_pdft_veff (mo, ci, incl_coul=True,
                paaa_only=True, state=ix)
            veff1.append (v1)
            veff2.append (v2)
        return mcpdft_grad.Gradients.kernel (self, state=state, mo=mo, ci=ci,
            si=si, d2f=d2f, veff1=veff1, veff2=veff2, _freeze_is=_freeze_is,
            **kwargs)

    def pack_uniq_var (self, xorb, xci, xis=None):
        x = sacasscf_grad.Gradients.pack_uniq_var (self, xorb, xci)
        if xis is not None: x = np.append (x, xis)
        return x

    def unpack_uniq_var (self, x):
        ngorb, nci, nis = self.ngorb, self.nci, self.nis
        x, xis = x[:ngorb+nci], x[ngorb+nci:]
        xorb, xci = sacasscf_grad.Gradients.unpack_uniq_var (self, x)
        if len (xis)==nis: return xorb, xci, xis
        return xorb, xci

    def _get_is_component (self, xci, ci=None, symm=-1):
        # TODO: state-average-mix
        if ci is None: ci = self.base.ci
        nroots = self.nroots
        xci = np.asarray (xci).reshape (nroots, -1)
        ci = np.asarray (ci).reshape (nroots, -1)
        xis = np.dot (xci.conj (), ci.T)
        if symm > -1: xis -= xis.T
        else:
            assert (np.amax (np.abs (xis + xis.T)) < 1e-8), '{}'.format (xis)
        return xis[np.tril_indices (nroots, k=-1)]

    def _separate_is_component (self, xci, ci=None, symm=-1):
        # TODO: state-average-mix
        is_list = isinstance (xci, list)
        is_tuple = isinstance (xci, tuple)
        if ci is None: ci = self.base.ci
        nroots = self.nroots
        ishape = np.asarray (xci).shape
        xci = np.asarray (xci).reshape (nroots, -1)
        xci = np.asarray (xci).reshape (nroots, -1)
        ci = np.asarray (ci).reshape (nroots, -1)
        xis = np.dot (xci.conj (), ci.T)
        xci -= np.dot (xis.conj (), ci)
        xci = xci.reshape (ishape)
        if is_list: xci = list (xci)
        elif is_tuple: xci = tuple (xci)
        if symm > -1: xis -= xis.T
        #else:
        #    assert (np.amax (np.abs (xis + xis.T)) < 1e-8), '{}'.format (xis)
        xis = xis[np.tril_indices (nroots, k=-1)]
        return xci, xis


    def get_wfn_response (self, si_bra=None, si_ket=None, state=None, mo=None,
            ci=None, si=None, eris=None, veff1=None, veff2=None,
            _freeze_is=False, d2f=None, **kwargs):
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if si is None: si = self.base.si
        if state is None: state = self.state
        ket, bra = _unpack_state (state)
        if si_bra is None: si_bra = si[:,bra]
        if si_ket is None: si_ket = si[:,ket]
        if d2f is None: d2f = self.base.diabatizer (ci=ci)[2]
        log = lib.logger.new_logger (self, self.verbose)
        si_diag = si_bra * si_ket

        # Diagonal: PDFT component
        nlag = self.nlag-self.nis
        g_all_pdft = np.zeros (nlag)
        for i, (amp, c, v1, v2) in enumerate (zip (si_diag, ci, veff1, veff2)):
            if not amp: continue
            g_i = mcpdft_grad.Gradients.get_wfn_response (self,
                state=i, mo=mo, ci=ci, veff1=v1, veff2=v2, nlag=nlag, **kwargs)
            g_all_pdft += amp * g_i
            if self.verbose >= lib.logger.DEBUG:
                g_orb, g_ci = self.unpack_uniq_var (g_i)
                g_ci, g_is = self._separate_is_component (g_ci, ci=ci, symm=0)
                log.debug ('g_is pdft state {} component:\n{} * {}'.format (i,
                    amp, g_is))

        # DEBUG
        g_orb_pdft, g_ci = self.unpack_uniq_var (g_all_pdft)
        g_ci, g_is_pdft = self._separate_is_component (g_ci, ci=ci, symm=0)

        # Off-diagonal: heff component
        g_orb_heff, g_is_heff = mspdft_heff_response (self, mo=mo, ci=ci,
            si_bra=si_bra, si_ket=si_ket, eris=eris)

        log.debug ('g_is pdft total component:\n{}'.format (g_is_pdft))
        log.debug ('g_is heff component:\n{}'.format (g_is_heff))

        # Combine
        g_orb = g_orb_pdft + g_orb_heff
        g_is = g_is_pdft + g_is_heff
        if _freeze_is: g_is[:] = 0.0
        g_all = self.pack_uniq_var (g_orb, g_ci, g_is)

        # DEBUG
        d2f_evals, d2f_evecs = linalg.eigh (d2f)
        g_is_modes = np.dot (g_is, d2f_evecs)
        g_is_pdft = np.dot (g_is_pdft, d2f_evecs)
        g_is_heff = np.dot (g_is_heff, d2f_evecs)
        log.debug ("IS sector Lagrange multiplier solution:")
        for i, (denom, num) in enumerate (zip (d2f_evals, g_is_modes)):
            log.debug ('%d %e / %e = %e (%e %e)', i, -num, denom, -num/denom,
                       g_is_pdft[i], g_is_heff[i])

        return g_all

    def get_Aop_Adiag (self, verbose=None, mo=None, ci=None, eris=None,
            level_shift=None, d2f=None, **kwargs):
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if eris is None and self.eris is None:
            eris = self.eris = self.base.ao2mo (mo)
        elif eris is None:
            eris = self.eris
        if d2f is None: d2f = self.base.diabatizer (ci=ci)[2]
        ham_od = self.base.get_heff_offdiag ()
        ham_od += ham_od.T # This corresponds to the arbitrary newton_casscf*2
        fcasscf = self.make_fcasscf_sa ()
        hop, Adiag = newton_casscf.gen_g_hop (fcasscf, mo, ci, eris,
            verbose)[2:]
        ngorb, nci = self.ngorb, self.nci
        # TODO: cacheing diab_response? or an x=0 branch?
        def Aop (x):
            x_v, x_is = x[:ngorb+nci], x[ngorb+nci:]
            Ax_v = hop (x_v) + self.diab_response (x_is, mo=mo, ci=ci,
                eris=eris)
            x_c = self.unpack_uniq_var (x_v)[1]
            Ax_is = np.dot (d2f, x_is)
            Ax_o, Ax_c = self.unpack_uniq_var (Ax_v)
            Ax_c, Ax_is2 = self._separate_is_component (Ax_c)
            Ax_c_od = list (np.tensordot (-ham_od, np.stack (x_c, axis=0),
                axes=1))
            Ax_c = [a1 + (w*a2) for a1, a2, w in zip (Ax_c, Ax_c_od,
                self.base.weights)]
            return self.pack_uniq_var (Ax_o, Ax_c, Ax_is)
        return Aop, Adiag

    def get_lagrange_precond (self, Adiag, level_shift=None, ci=None, d2f=None,
            **kwargs):
        if level_shift is None: level_shift = self.level_shift
        if ci is None: ci = self.base.ci
        if d2f is None: d2f = self.base.diabatizer (ci=ci)[2]
        return MSPDFTLagPrec (Adiag=Adiag, level_shift=level_shift, ci=ci,
            d2f=d2f, grad_method=self)

    def get_ham_response (self, si_bra=None, si_ket=None, state=None, mo=None,
            ci=None, si=None, eris=None, veff1=None, veff2=None, mf_grad=None,
            atmlst=None, verbose=None, **kwargs):
        '''write mspdft heff Hellmann-Feynman calculator; sum over
        diagonal PDFT Hellmann-Feynman terms
        '''
        if atmlst is None: atmlst = self.atmlst
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if si is None: si = self.base.si
        if state is None: state = self.state
        ket, bra = _unpack_state (state)
        if si_bra is None: si_bra = si[:,bra]
        if si_ket is None: si_ket = si[:,ket]
        if mf_grad is None: mf_grad = self.base.get_rhf_base ().nuc_grad_method ()
        if verbose is None: verbose = self.verbose
        si_diag = si_bra * si_ket
        log = logger.new_logger (self, verbose)

        # Fix messed-up counting of the nuclear part
        de_nuc = mf_grad.grad_nuc (self.mol, atmlst)
        log.debug ('MS-PDFT gradient n-n terms:\n{}'.format (de_nuc))
        de = si_diag.sum () * de_nuc.copy ()

        # Diagonal: PDFT component
        for i, (amp, c, v1, v2) in enumerate (zip (si_diag, ci, veff1, veff2)):
            if not amp: continue
            de_i = mcpdft_grad.Gradients.get_ham_response (self, state=i,
                mo=mo, ci=ci, veff1=v1, veff2=v2, eris=eris, mf_grad=mf_grad,
                verbose=0, **kwargs) - de_nuc
            log.debug ('MS-PDFT gradient int-state {} EPDFT terms:\n{}'.format
                (i, de_i))
            log.debug ('Factor for these terms: {}'.format (amp))
            de += amp * de_i
        log.debug ('MS-PDFT gradient diag H-F terms:\n{}'.format (de))

        # Off-diagonal: heff component
        de_o = mspdft_heff_HellmanFeynman (self, mo_coeff=mo, ci=ci,
            si_bra=si_bra, si_ket=si_ket, eris=eris, state=state,
            mf_grad=mf_grad, **kwargs)
        log.debug ('MS-PDFT gradient offdiag H-F terms:\n{}'.format (de_o))
        de += de_o

        return de

    def get_LdotJnuc (self, Lvec, atmlst=None, verbose=None, mo=None,
            ci=None, eris=None, mf_grad=None, d2f=None, **kwargs):
        ''' Add the IS component '''
        if atmlst is None: atmlst = self.atmlst
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if eris is None and self.eris is None:
            eris = self.eris = self.base.ao2mo (mo)
        elif eris is None:
            eris = self.eris

        ngorb, nci = self.ngorb, self.nci
        Lvec_v, Lvec_is = Lvec[:ngorb+nci], Lvec[ngorb+nci:]

        # Double-check Lvec_v sanity
        Lvec_orb, Lvec_ci = self.unpack_uniq_var (Lvec_v)
        Lvec_is2 = self._get_is_component (Lvec_ci, symm=0)
        assert (np.amax (np.abs (Lvec_is2)) < 1e-8), '{} {}'.format (Lvec_is,
            Lvec_is2)

        # Orbital and CI components
        de_Lv = sacasscf_grad.Gradients.get_LdotJnuc (self, Lvec_v,
            atmlst=atmlst, verbose=verbose, ci=ci, eris=eris, mf_grad=mf_grad,
            **kwargs)

        # SI component
        t0 = (logger.process_clock(), logger.perf_counter())
        de_Lis = self.diab_grad (Lvec_is, atmlst=atmlst, mf_grad=mf_grad,
            eris=eris, mo=mo, ci=ci, **kwargs)
        logger.info (self,
            '--------------- %s gradient Lagrange IS response ---------------',
            self.base.__class__.__name__)
        if verbose >= logger.INFO:
            rhf_grad._write(self, self.mol, de_Lis, atmlst)
        logger.info (self,
            '----------------------------------------------------------------')
        t0 = logger.timer (self, '{} gradient Lagrange IS response'.format (
            self.base.__class__.__name__), *t0)
        return de_Lv + de_Lis

    def get_lagrange_callback (self, Lvec_last, itvec, geff_op):
        # TODO: state-average-mix
        log = logger.new_logger (self, self.verbose)
        if isinstance (self.base.fcisolver, CSFFCISolver):
            transf = self.base.fcisolver.transformer
            def _debug_csfs (xci, tag):
                xci_csf = transf.vec_det2csf (xci, normalize=False)
                xci_p = transf.vec_csf2det (xci_csf, normalize=False)
                xci_bs_norm = linalg.norm (np.concatenate (
                    [x.ravel () - y.ravel () for x, y in zip (xci_p, xci)]))
                log.debug ('Broken-spin |{}| = {}'.format (tag, xci_bs_norm))
        else:
            def _debug_csfs (xci, tag):
                pass
        def my_call (x):
            itvec[0] += 1
            geff = geff_op (x)
            deltax = x - Lvec_last
            gorb, gci, gis = self.unpack_uniq_var (geff)
            deltaorb, deltaci, deltais = self.unpack_uniq_var (deltax)
            gci_norm = linalg.norm (np.asarray (gci).ravel ())
            deltaci_norm = linalg.norm (np.asarray (deltaci).ravel ())
            logger.info(self, ('Lagrange optimization iteration {}, |gorb| = '
                '{}, |gci| = {}, |gis| = {} |dLorb| = {}, |dLci| = {}, |dLis|'
                ' = {}').format (itvec[0], linalg.norm (gorb),
                gci_norm, linalg.norm (gis), linalg.norm (deltaorb),
                deltaci_norm, linalg.norm (deltais)))
            _debug_csfs (gci, 'gci')
            _debug_csfs (deltaci, 'dLci')
            Lvec_last[:] = x[:]
        return my_call

    def debug_lagrange (self, Lvec, bvec, Aop, Adiag, state=None, mo=None,
            ci=None, d2f=None, verbose=None, eris=None, **kwargs):
        if state is None: state = self.state
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if eris is None: eris = self.base.ao2mo (mo)
        if verbose is None: verbose = self.verbose
        nroots = self.nroots
        log = logger.new_logger (self, verbose)
        if isinstance (self.base.fcisolver, CSFFCISolver):
            transf = self.base.fcisolver.transformer
            def _debug_csfs (xci, label, normalize=False):
                strs, vecs = transf.printable_largest_csf (np.asarray (xci),
                    10, isdet=True, normalize=normalize, order='C')
                log.debug ('Leading CSFs for %s', label)
                for iroot in range (nroots):
                    log.debug (' Root %d', iroot)
                    for s, v in zip (strs[iroot], vecs[iroot]):
                        log.debug ('  |%s>: %s', s, v)
        else:
            def _debug_csfs (*args, **kwargs):
                pass
        def _debug_cispace (xci, label):
            log.debug ('%s', label)
            xci_norm = [np.dot (c.ravel (), c.ravel ()) for c in xci]
            try:
                xci_ss = self.base.fcisolver.states_spin_square (xci,
                    self.base.ncas, self.base.nelecas)[0]
            except AttributeError:
                from pyscf.fci.direct_spin1 import _unpack_nelec, spin_square
                nelec = sum (_unpack_nelec(self.base.nelecas))
                xci_ss = [spin_square (x, self.base.ncas,
                          ((nelec+m)//2,(nelec-m)//2))[0]
                          for x, m in zip (xci, self.spin_states)]

            xci_ss = [x / max (y, 1e-8) for x, y in zip (xci_ss, xci_norm)]
            xci_multip = [np.sqrt (x+.25) - .5 for x in xci_ss]
            xci_norm = np.sqrt (xci_norm)
            for ix, (norm, ss, multip) in enumerate (zip (xci_norm, xci_ss,
                    xci_multip)):
                log.debug ((' State {} norm = {:.7e} ; <S^2> = {:.7f} ; 2S+1'
                            ' = {:.7f}').format (ix, norm, ss, multip))
            ovlp = np.zeros ((nroots, nroots), dtype=xci[0].dtype)
            for i, j in product (range (nroots), repeat=2):
                if self.spin_states[i] != self.spin_states[j]: continue
                ovlp[i,j] = np.dot (xci[i].ravel (), ci[j].ravel ())
            log.debug (' Overlap matrix with CI array:')
            fmt_str = '  ' + ' '.join (['{:8.1e}',]*nroots)
            for row in ovlp: log.debug (fmt_str.format (*row))
        _debug_csfs (ci, 'CI vector', normalize=True)
        borb, bci, bis = self.unpack_uniq_var (bvec)
        log.debug ('Orbital rotation gradient (b) norm = {:.6e}'.format (
            linalg.norm (borb)))
        _debug_cispace (bci, 'CI gradient (b)')
        _debug_csfs (bci, 'CI gradient (b)')
        Aorb, Aci = self.unpack_uniq_var (Adiag)
        log.debug ('Orbital rotation Hessian (A) diagonal norm = {:.7e}'.format
            (linalg.norm (Aorb)))
        _debug_cispace (Aci, 'CI Hessian (A) diagonal')
        _debug_csfs (Aci, 'CI Hessian (A) diagonal')
        Lorb, Lci, Lis = self.unpack_uniq_var (Lvec)
        log.debug ('Orbital rotation Lagrange vector (x) norm = {:.7e}'.format
            (linalg.norm (Lorb)))
        _debug_cispace (Lci, 'CI Lagrange (x) vector')
        _debug_csfs (Lci, 'CI Lagrange (x) vector')
        log.debug ('{} Constraint Jacobian (A):'.format (self.base.__class__.__name__))
        fmt = ' ' + ' '.join (['{:12.5e}' for i in range (self.nis)])
        for row in d2f: log.debug (fmt.format (*row))
        log.debug (' {:>12s} {:>12s}'.format ('Gradient (b)', 'Vector (x)'))
        for g, v in zip (bis, Lis):
            log.debug (' {:12.5e} {:12.5e}'.format (g, v))

class MSPDFTLagPrec (sacasscf_grad.SACASLagPrec):
    ''' Solve IS part exactly, then do everything else the same '''

    def __init__(self, Adiag=None, level_shift=None, ci=None, grad_method=None,
            d2f=None, **kwargs):
        sacasscf_grad.SACASLagPrec.__init__(self, Adiag=Adiag,
            level_shift=level_shift, ci=ci, grad_method=grad_method)
        self.grad_method = grad_method
        self.sing_tol = getattr (grad_method.base, 'sing_tol_diabatize',
                                 SING_TOL_DIABATIZE)
        self.sing_step_tol = getattr (grad_method.base, 'sing_step_tol',
                                      SING_STEP_TOL)
        self.sing_warned = False
        self.log = logger.new_logger (self.grad_method,
            self.grad_method.verbose)
        self._init_d2f (d2f=d2f, **kwargs)
        self.verbose = self.grad_method.verbose

    def _init_d2f (self, d2f=None, **kwargs):
        self.d2f=d2f
        self.d2f_evals, self.d2f_evecs = linalg.eigh (d2f)
        idx_sing = np.abs (self.d2f_evals) < self.sing_tol
        self.log.debug ('IS component Hessian eigenvalues: {}'.format (
            self.d2f_evals))
        if np.any (idx_sing): self.do_sing_warn ()
        self.d2f_evals = self.d2f_evals[~idx_sing]
        self.d2f_evecs = self.d2f_evecs[:,~idx_sing]

    def unpack_uniq_var (self, x):
        return self.grad_method.unpack_uniq_var (x)

    def pack_uniq_var (self, x0, x1, x2=None):
        return self.grad_method.pack_uniq_var (x0, x1, x2)

    def __call__(self, x):
        xorb, xci, xis = self.unpack_uniq_var (x)
        Mxorb = self.orb_prec (xorb)
        Mxci = self.ci_prec (xci)
        Mxis = self.is_prec (xis)
        return self.pack_uniq_var (Mxorb, Mxci, Mxis)

    def is_prec (self, xis):
        xis = np.dot (xis, self.d2f_evecs)
        Mxis = xis/self.d2f_evals
        idx_sing = (np.abs (Mxis) >= self.sing_step_tol)
        if np.any (idx_sing): self.do_sing_warn ()
        Mxis[idx_sing] = 0
        Mxis = np.dot (self.d2f_evecs, Mxis)
        return Mxis

    def do_sing_warn (self):
        if self.sing_warned: return
        self.log.warn ('Model-space frame-rotation Hessian is singular! '
                        'Response equations may not be solvable to arbitrary '
                        'precision!')
        self.sing_warned = True



if __name__ == '__main__':
    # Test mspdft_heff_response and mspdft_heff_HellmannFeynman by trying to
    # reproduce SA-CASSCF derivatives in an arbitrary basis
    import math
    from pyscf import scf, gto, mcscf
    from pyscf.fci import csf_solver
    xyz = '''O  0.00000000   0.08111156   0.00000000
             H  0.78620605   0.66349738   0.00000000
             H -0.78620605   0.66349738   0.00000000'''
    mol = gto.M (atom=xyz, basis='6-31g', symmetry=False, output='mspdft.log',
        verbose=lib.logger.DEBUG)
    mf = scf.RHF (mol).run ()
    mc = mcpdft.CASSCF (mf, 'tPBE', 4, 4).set (fcisolver = csf_solver (mol, 1))
    mc = mc.multi_state ([1.0/3,]*3, 'cms').run ()
    mc_grad = Gradients (mc)
    de = np.stack ([mc_grad.kernel (state=i) for i in range (3)], axis=0)




#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Matthew Hennefarth <mhennefarth@uchicago.com>

import numpy as np
from pyscf.lib import logger, tag_array, current_memory
from pyscf.dft.gen_grid import BLKSIZE
from pyscf.mcpdft.otpd import get_ontop_pair_density
from pyscf.mcpdft.pdft_eff import _ERIS
from pyscf.mcpdft.tfnal_derivs import contract_fot
import gc


def kernel(ot, dm1s, cascm2, c_dm1s, c_cascm2, mo_coeff, ncore, ncas, max_memory=2000, hermi=1, paaa_only=False,
           aaaa_only=False, jk_pc=False, delta=False):
    r'''Get the 1- and 2-body effective gradient responses from MC-PDFT. The
    $\rho \cdot \mathbf{F}$ terms, or Hessian vector products.

    Args:
        ot : an instance of otfnal class
        dm1s : ndarray of shape (2, nao, nao)
            Contains the spin-separated one-body density matrices to evaluate
            the kernel at
        cascm2 : ndarray of shape (ncas, ncas, ncas, ncas)
            Spin-summed two-body cumulant density matrix in the active space to
            evaluate the kernel at
        c_dm1s : ndarray of shape (2, nao, nao)
            Contains the spin-separated one-body density matrices to contract
            the kernel with.
        c_cascm2 : ndarray of shape (ncas, ncas, ncas, ncas)
            Spin-summed two-body cumulant density matrix in the active space to
            contract the kernel with.
        mo_coeff : ndarray of shape (nao, nmo)
            containing molecular orbital coefficients
        ncore : integer
            number of inactive orbitals
        ncas : integer
            number of active orbitals

    Kwargs:
        max_memory : int or float
            maximum cache size in MB
            default is 2000
        hermi : int
            1 if 1rdms are assumed hermitian, 0 otherwise
        paaa_only : logical
            If true, only compute the paaa range of papa and ppaa (all other
            elements set to zero)
        aaaa_only : logical
            If true, only compute the aaaa range of papa and ppaa (all other
            elements set to zero; overrides paaa_only)
        jk_pc : logical
            If true, compute the ppii=pipi elements of veff2 (otherwise, these
            are set to zero)
        delta : logical
            If true, then contract with the delta density. The delta density is the c_dm1s-dm1s and similarly for the
            2rdm element (though care is taken since 2rdm elements are expressed in cumulant form).

    Returns:
        feff1 : ndarray of shape (nao, nao)
            1-body effective gradient response
        feff2 : object of class pdft_eff._ERIS
            2-body effective gradient response
    '''
    nocc = ncore + ncas
    ni, xctype, dens_deriv = ot._numint, ot.xctype, ot.dens_deriv
    nao = mo_coeff.shape[0]
    mo_core = mo_coeff[:, :ncore]
    mo_cas = mo_coeff[:, ncore:nocc]
    shls_slice = (0, ot.mol.nbas)
    ao_loc = ot.mol.ao_loc_nr()

    omega, alpha, hyb = ot._numint.rsh_and_hybrid_coeff(ot.otxc)
    hyb_x, hyb_c = hyb
    if abs(omega) > 1e-11:
        raise NotImplementedError("range-separated on-top functionals")

    if abs(hyb_x) > 1e-11 or abs(hyb_c) > 1e-11:
        raise NotImplementedError("effective potential for hybrid functionals")

    feff1 = np.zeros((nao, nao), dtype=dm1s.dtype)
    feff2 = _ERIS(ot.mol, mo_coeff, ncore, ncas, paaa_only=paaa_only,
                  aaaa_only=aaaa_only, jk_pc=jk_pc, verbose=ot.verbose,
                  stdout=ot.stdout)

    t0 = (logger.process_clock(), logger.perf_counter())

    # Density matrices
    dm_core = mo_core @ mo_core.T
    dm_cas = dm1s - dm_core[None, :, :]
    dm_core *= 2

    # Propagate speedup tags
    if hasattr(dm1s, 'mo_coeff') and hasattr(dm1s, 'mo_occ'):
        dm_core = tag_array(dm_core, mo_coeff=dm1s.mo_coeff[0, :, :ncore],
                            mo_occ=dm1s.mo_occ[:, :ncore].sum(0))
        dm_cas = tag_array(dm_cas, mo_coeff=dm1s.mo_coeff[:, :, ncore:nocc],
                           mo_occ=dm1s.mo_occ[:, ncore:nocc])

    # rho generators
    make_rho_c = ni._gen_rho_evaluator(ot.mol, dm_core, hermi)[0]
    make_rho_a = ni._gen_rho_evaluator(ot.mol, dm_cas, hermi)[0]
    make_rho = ni._gen_rho_evaluator(ot.mol, dm1s, hermi)[0]
    make_crho = ni._gen_rho_evaluator(ot.mol, c_dm1s, hermi)[0]

    # memory block size
    gc.collect()
    remaining_floats = (max_memory - current_memory()[0]) * 1e6 / 8
    nderiv_rho = (1, 4, 10)[dens_deriv]  # ?? for meta-GGA
    nderiv_Pi = (1, 4)[ot.Pi_deriv]
    ncols = 4 + nderiv_rho * nao  # ao, weight, coords
    ncols += nderiv_rho * 4 + nderiv_Pi  # rho, rho_a, rho_c, Pi
    ncols += 1 + nderiv_rho + nderiv_Pi  # eot, vot
    ncols += (nderiv_rho*nderiv_Pi*(nderiv_rho*nderiv_Pi + 1))/2 # fot

    # Asynchronous part
    nfeff1 = nderiv_rho * (nao + 1)  # footprint of get_eff_1body
    nfeff2 = feff2._accumulate_ftpt() * nderiv_Pi
    ncols += np.amax([nfeff1, nfeff2])  # asynchronous fns
    pdft_blksize = int(remaining_floats / (ncols * BLKSIZE)) * BLKSIZE
    if ot.grids.coords is None:
        ot.grids.build(with_non0tab=True)
    ngrids = ot.grids.coords.shape[0]
    ngrids_blk = int(ngrids / BLKSIZE) * BLKSIZE
    pdft_blksize = max(BLKSIZE, min(pdft_blksize, ngrids_blk, BLKSIZE * 1200))
    logger.debug(ot, ('{} MB used of {} available; block size of {} chosen'
        'for grid with {} points').format(current_memory()[0], max_memory,
                                          pdft_blksize, ngrids))

    for ao, mask, weight, coords in ni.block_loop(ot.mol, ot.grids, nao,
                                                  dens_deriv, max_memory,
                                                  blksize=pdft_blksize):
        rho = np.asarray([make_rho(i, ao, mask, xctype) for i in range(2)])
        crho = np.asarray([make_crho(i, ao, mask, xctype) for i in range(2)])
        rho_a = sum([make_rho_a(i, ao, mask, xctype) for i in range(2)])
        rho_c = make_rho_c(0, ao, mask, xctype)
        t0 = logger.timer(ot, 'untransformed densities (core and total)', *t0)

        Pi = get_ontop_pair_density(ot, rho, ao, cascm2, mo_cas,
                                    dens_deriv, mask)
        cPi = get_ontop_pair_density(ot, crho, ao, c_cascm2, mo_cas,
                                     dens_deriv, mask)
        t0 = logger.timer(ot, 'on-top pair density calculation', *t0)

        if delta:
            crho -= rho
            cPi -= Pi

        vot, fot = ot.eval_ot(rho, Pi, weights=weight, dderiv=2,
                              _unpack_vot=False)[1:]
        frho, fPi = contract_fot(ot, fot, rho, Pi, crho, cPi, unpack=True,
                                 vot_packed=vot)

        t0 = logger.timer(ot, 'effective gradient response kernel calculation',
                          *t0)

        if ao.ndim == 2:
            ao = ao[None, :, :]

        feff1 += ot.get_eff_1body(ao, weight, frho, non0tab=mask,
                                  shls_slice=shls_slice, ao_loc=ao_loc,
                                  hermi=1)
        t0 = logger.timer(ot, '1-body effective gradient response calculation',
                          *t0)

        feff2._accumulate(ot, ao, weight, rho_c, rho_a, fPi, mask, shls_slice,
                          ao_loc)
        t0 = logger.timer(ot, '2-body effective gradient response calculation',
                          *t0)

    feff2._finalize()
    t0 = logger.timer(ot, 'Finalizing 2-body gradient response calculation',
                      *t0)

    return feff1, feff2


def lazy_kernel(ot, dm1s, cascm2, c_dm1s, c_cascm2, mo_cas, hermi=1, max_memory=2000, delta=False):
    '''1- and 2-body gradient response (hessian-vector products) from MC-PDFT.
    This is the lazy way and doesn't care about memory.'''
    ni, xctype, dens_deriv = ot._numint, ot.xctype, ot.dens_deriv
    nao = mo_cas.shape[0]

    feff1 = np.zeros_like(dm1s[0])
    feff2 = np.zeros((nao, nao, nao, nao), dtype=feff1.dtype)

    t0 = (logger.process_clock(), logger.perf_counter())

    make_rho = tuple(ni._gen_rho_evaluator(ot.mol, dm1s[i, :, :], hermi) for i
        in range(2))
    make_crho = tuple(ni._gen_rho_evaluator(ot.mol, c_dm1s[i, :, :], hermi) for
        i in range(2))

    for ao, mask, weight, coords in ni.block_loop(ot.mol, ot.grids, nao,
                                                  dens_deriv, max_memory):
        rho = np.asarray([m[0](0, ao, mask, xctype) for m in make_rho])
        crho = np.asarray([m[0](0, ao, mask, xctype) for m in make_crho])
        t0 = logger.timer(ot, 'untransformed density', *t0)
        Pi = get_ontop_pair_density(ot, rho, ao, cascm2, mo_cas, dens_deriv,
                                    mask)
        cPi = get_ontop_pair_density(ot, crho, ao, c_cascm2, mo_cas,
                                     dens_deriv, mask)
        t0 = logger.timer(ot, 'on-top pair density calculation', *t0)

        if delta:
            crho -= rho
            cPi -= Pi

        vot, fot = ot.eval_ot(rho, Pi, weights=weight, dderiv=2,
                              _unpack_vot=False)[1:]
        frho, fPi = contract_fot(ot, fot, rho, Pi, crho, cPi, unpack=True,
                                 vot_packed=vot)

        t0 = logger.timer(ot, 'effective gradient response kernel calculation',
                          *t0)
        if ao.ndim == 2:
            ao = ao[None, :, :]

        feff1 += ot.get_eff_1body(ao, weight, frho)
        t0 = logger.timer(ot, '1-body effective gradient response calculation',
                          *t0)

        feff2 += ot.get_eff_2body(ao, weight, fPi, aosym=1)
        t0 = logger.timer(ot, '2-body effective gradient response calculation',
                          *t0)

    return feff1, feff2


def get_feff_1body(otfnal, rho, Pi, crho, cPi, ao, weight, kern=None, non0tab=None,
                   shls_slice=None, ao_loc=None, hermi=0, **kwargs):
    r"""Get the terms [\Delta F]_{pq}

    Args:
        rho : ndarray of shape (2,*,ngrids)
            Spin-density [and derivatives]
        Pi : ndarray with shape (*,ngrids)
            On-top pair density [and derivatives]
        crho : ndarray of shape (2,*,ngrids)
            Spin-density [and derivatives] to contract the hessian with
        cPi : ndarray with shape (*,ngrids)
            On-top pair density [and derivatives] to contract Hessian with
        ao : ndarray or 2 ndarrays of shape (*,ngrids,nao)
            contains values and derivatives of nao. 2 different ndarrays can
            have different nao but not different ngrids
        weight : ndarray of shape (ngrids)
            containing numerical integration weights

    Kwargs:
        kern : ndarray of shape (*,ngrids)
            the hessian-vector product. If not provided, it is calculated.
        non0tab : ndarray of shape (nblk, nbas)
            Identifies blocks of grid points which are nonzero on each AO shell
            so as to exploit sparsity. If you want the "ao" array to be in the
            MO basis, just leave this as None. If hermi == 0, it only applies
            to the bra index ao array, even if the ket index ao array is the
            same (so probably always pass hermi = 1 in that case)
        shls_slice : sequence of integers of len 2
            Identifies starting and stopping indices of AO shells
        ao_loc : ndarray of length nbas
            Offset to first AO of each shell
        hermi : integer or logical
            Toggle whether feff is supposed to be a Hermitian matrix You can
            still pass two different ao arrays for the bra and the ket indices,
            for instance if one of them is supposed to be a higher derivative.
            They just have to have the same nao in that case.

    Returns : ndarray of shape (nao[0],nao[1])
        The 1-body effective gradient response corresponding to this on-top
        pair density exchange-correlation functional, in the atomic-orbital
        basis. In PDFT this functional is always spin-symmetric.
    """
    if kern is None:
        if rho.ndim == 2:
            rho = np.expand_dims(rho, 1)
            Pi = np.expand_dims(Pi, 0)

        if crho.ndim == 2:
            crho = np.expand_dims(crho, 1)
            cPi = np.expand_dims(cPi, 0)

        fot = otfnal.eval_ot(rho, Pi, dderiv=2, **kwargs)[2]
        kern = contract_fot(otfnal, fot, rho, Pi, crho, cPi)[0]
        rho = np.squeeze(rho)
        Pi = np.squeeze(Pi)
        crho = np.squeeze(crho)
        cPi = np.squeeze(cPi)

    return otfnal.get_eff_1body(ao, weight, kern=kern, non0tab=non0tab,
                                shls_slice=shls_slice, ao_loc=ao_loc,
                                hermi=hermi)


def get_feff_2body(otfnal, rho, Pi, crho, cPi, ao, weight, aosym='s4',
                   kern=None, fao=None, **kwargs):
    r'''Get the terms [\Delta F]_{pqrs}

        Args:
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]
            crho : ndarray of shape (2,*,ngrids)
                Spin-density [and derivatives] to contract the hessian with
            cPi : ndarray with shape (*,ngrids)
                On-top pair density [and derivatives] to contract Hessian with
            ao : ndarray of shape (*,ngrids,nao)
                OR list of ndarrays of shape (*,ngrids,*) values and
                derivatives of atomic or molecular orbitals in which space to
                calculate the 2-body veff If a list of length 4, the
                corresponding set of eri-like elements are returned
            weight : ndarray of shape (ngrids)
                containing numerical integration weights

        Kwargs:
            aosym : int or str
                Index permutation symmetry of the desired integrals. Valid
                options are 1 (or '1' or 's1'), 4 (or '4' or 's4'), '2ij' (or
                's2ij'), and '2kl' (or 's2kl'). These have the same meaning as
                in PySCF's ao2mo module. Currently all symmetry exploitation is
                extremely slow and unparallelizable for some reason so trying
                to use this is not recommended until I come up with a C
                routine.
            kern : ndarray of shape (*,ngrids)
                the hessian-vector product. If not provided, it is calculated.
            fao : ndarray of shape (*,ngrids,nao,nao) or
                (*,ngrids,nao*(nao+1)//2). An intermediate in which the kernel
                and the k,l orbital indices have been contracted. Overrides
                kl_symm

        Returns : eri-like ndarray
            The two-body effective gradient response corresponding to this on-top
            pair density exchange-correlation functional or elements
            thereof, in the provided basis.
        '''
    if kern is None:
        if rho.ndim == 2:
            rho = np.expand_dims(rho, 1)
            Pi = np.expand_dims(Pi, 0)

        if crho.ndim == 2:
            crho = np.expand_dims(crho, 1)
            cPi = np.expand_dims(cPi, 0)

        fot = otfnal.eval_ot(rho, Pi, dderiv=2, **kwargs)[2]
        kern = contract_fot(otfnal, fot, rho, Pi, crho, cPi)[1]
        rho = np.squeeze(rho)
        Pi = np.squeeze(Pi)
        crho = np.squeeze(crho)
        cPi = np.squeeze(cPi)

    return otfnal.get_eff_2body(ao, weight, kern, aosym=aosym, eff_ao=fao)


def get_feff_2body_kl(otfnal, rho, Pi, crho, cPi, ao_k, ao_l, weight,
                      symm=False, kern=None, **kwargs):
    r''' get the two-index intermediate Mkl of [\Delta \cdot F]_{pqrs}

        Args:
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]
            crho : ndarray of shape (2,*,ngrids)
                Spin-density [and derivatives] to contract the hessian with
            cPi : ndarray with shape (*,ngrids)
                On-top pair density [and derivatives] to contract Hessian with
            ao_k : ndarray of shape (*,ngrids,nao)
                OR list of ndarrays of shape (*,ngrids,*) values and
                derivatives of atomic or molecular orbitals corresponding to
                index k
            ao_l : ndarray of shape (*,ngrids,nao)
                OR list of ndarrays of shape (*,ngrids,*) values and
                derivatives of atomic or molecular orbitals corresponding to
                index l
            weight : ndarray of shape (ngrids)
                containing numerical integration weights

        Kwargs:
            symm : logical
                Index permutation symmetry of the desired integral wrt k,l
            kern : ndarray of shape (*,ngrids)
                the hessian-vector product. If not provided, it is calculated.

        Returns : ndarray of shape (*,ngrids,nao,nao)
            or (*,ngrids,nao*(nao+1)//2). An intermediate for calculating the
            two-body effective gradient response corresponding to this on-top
            pair density exchange-correlation functional in the provided basis.
        '''
    if kern is None:
        if rho.ndim == 2:
            rho = np.expand_dims(rho, 1)
            Pi = np.expand_dims(Pi, 0)

        if crho.ndim == 2:
            crho = np.expand_dims(crho, 1)
            cPi = np.expand_dims(cPi, 0)

        fot = otfnal.eval_ot(rho, Pi, dderiv=2, **kwargs)[2]
        kern = contract_fot(otfnal, fot, rho, Pi, crho, cPi)[1]

    return otfnal.get_eff_2body_kl(ao_k, ao_l, weight, kern=kern, symm=symm)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
4-component Dirac-Kohn-Sham
'''


import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import dhf
from pyscf.dft import rks
from pyscf.dft import gks
from pyscf.dft import r_numint


@lib.with_doc(gks.get_veff.__doc__)
def get_veff(ks, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
    if ks.do_nlc():
        raise NotImplementedError(f'NLC functional {ks.xc} + {ks.nlc}')
    return gks.get_veff(ks, mol, dm, dm_last, vhf_last, hermi)


def energy_elec(ks, dm=None, h1e=None, vhf=None):
    r'''Electronic part of DKS energy.

    Note this function has side effects which cause mf.scf_summary updated.

    Args:
        ks : an instance of DFT class

        dm : 2D ndarray
            one-particle density matrix
        h1e : 2D ndarray
            Core hamiltonian

    Returns:
        DKS electronic energy and the 2-electron contribution
    '''
    e1, e2 = rks.energy_elec(ks, dm, h1e, vhf)
    if not ks.with_ssss and ks.ssss_approx == 'Visscher':
        e2 += dhf._vischer_ssss_correction(ks, dm)
        ks.scf_summary['e2'] = e2
    return e1, e2


class KohnShamDFT(rks.KohnShamDFT):
    def __init__(self, xc='LDA,VWN'):
        rks.KohnShamDFT.__init__(self, xc)
        self._numint = r_numint.RNumInt()

    def dump_flags(self, verbose=None):
        rks.KohnShamDFT.dump_flags(self, verbose)
        logger.info(self, 'collinear = %s', self._numint.collinear)
        if self._numint.collinear[0] == 'm':
            logger.info(self, 'mcfun spin_samples = %s', self._numint.spin_samples)
            logger.info(self, 'mcfun collinear_thrd = %s', self._numint.collinear_thrd)
            logger.info(self, 'mcfun collinear_samples = %s', self._numint.collinear_samples)
        return self

    get_veff = gks.get_veff
    energy_elec = gks.energy_elec

    @property
    def collinear(self):
        return self._numint.collinear
    @collinear.setter
    def collinear(self, val):
        self._numint.collinear = val

    @property
    def spin_samples(self):
        return self._numint.spin_samples
    @spin_samples.setter
    def spin_samples(self, val):
        self._numint.spin_samples = val

    def to_rhf(self):
        raise RuntimeError

    def to_uhf(self):
        raise RuntimeError

    def to_ghf(self):
        raise RuntimeError

    def to_rks(self, xc=None):
        raise RuntimeError

    def to_uks(self, xc=None):
        raise RuntimeError

    def to_gks(self, xc=None):
        raise RuntimeError

    def to_dhf(self):
        '''Convert the input mean-field object to a DHF object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        mf = self.view(dhf.DHF)
        mf.converged = False
        return mf

    to_hf = to_dhf

    def to_dks(self, xc=None):
        if xc is not None and xc != self.xc:
            mf = self.copy()
            mf.xc = xc
            mf.converged = False
        return self

    to_ks = to_dks


class DKS(KohnShamDFT, dhf.DHF):
    '''Kramers unrestricted Dirac-Kohn-Sham'''
    def __init__(self, mol, xc='LDA,VWN'):
        dhf.DHF.__init__(self, mol)
        KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        dhf.DHF.dump_flags(self, verbose)
        KohnShamDFT.dump_flags(self, verbose)
        return self

    def x2c1e(self):
        from pyscf.x2c import dft
        x2chf = dft.UKS(self.mol)
        x2chf.__dict__.update(self.__dict__)
        return x2chf
    x2c = x2c1e

    to_gpu = lib.to_gpu

UKS = UDKS = DKS

class RDKS(DKS, dhf.RDHF):
    '''Kramers restricted Dirac-Kohn-Sham'''
    def __init__(self, mol, xc='LDA,VWN'):
        dhf.RDHF.__init__(self, mol)
        KohnShamDFT.__init__(self, xc)

    def x2c1e(self):
        from pyscf.x2c import dft
        x2chf = dft.RKS(self.mol)
        x2chf.__dict__.update(self.__dict__)
        return x2chf
    x2c = x2c1e

    def to_dhf(self):
        '''Convert the input mean-field object to a DHF object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        mf = self.view(dhf.RDHF)
        mf.converged = False
        return mf

RKS = RDKS



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
density fitting MP2,  3-center integrals incore.
'''

import h5py
import ctypes
import numpy as np
import scipy.linalg
from functools import reduce
from pyscf import lib
from pyscf.lib import logger
from pyscf.ao2mo import _ao2mo
from pyscf import df
from pyscf.mp import dfmp2, ump2
from pyscf.mp.ump2 import make_rdm1, make_rdm2, _mo_splitter
from pyscf import __config__

WITH_T2 = getattr(__config__, 'mp_dfump2_with_t2', True)
THRESH_LINDEP = getattr(__config__, 'mp_dfump2_thresh_lindep', 1e-10)

libmp = dfmp2.libmp


def kernel(mp, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2, verbose=None):

    log = logger.new_logger(mp, verbose)

    if eris is None: eris = mp.ao2mo()

    nocc, nvir, naux = eris.nocc, eris.nvir, eris.naux
    nvirmax = max(nvir)
    split_mo_energy = mp.split_mo_energy()
    occ_energy = [x[1] for x in split_mo_energy]
    vir_energy = [x[2] for x in split_mo_energy]

    mem_avail = mp.max_memory - lib.current_memory()[0]

    if with_t2:
        t2 = (np.zeros((nocc[0],nocc[0],nvir[0],nvir[0]), dtype=eris.dtype),
              np.zeros((nocc[0],nocc[1],nvir[0],nvir[1]), dtype=eris.dtype),
              np.zeros((nocc[1],nocc[1],nvir[1],nvir[1]), dtype=eris.dtype))
        t2_ptr = [x.ctypes.data_as(ctypes.c_void_p) for x in t2]
        mem_avail -= sum([x.size for x in t2]) * eris.dsize / 1e6
    else:
        t2 = None
        t2_ptr = [lib.c_null_ptr()] * 3

    if mem_avail < 0:
        log.error('Insufficient memory for holding t2 incore. Please rerun with `with_t2 = False`.')
        raise MemoryError

    emp2_ss = emp2_os = 0

    drv = libmp.MP2_contract_d

    # determine occ blksize
    if isinstance(eris.ovL[0], np.ndarray):    # incore ovL
        occ_blksize = nocc
    else:   # outcore ovL
        # 3*V^2 (for C driver) + 2*[O]XV (for iaL & jaL) = mem
        occ_blksize = int(np.floor((mem_avail*0.6*1e6/eris.dsize - 3*nvirmax**2)/(2*naux*nvirmax)))
        occ_blksize = [min(nocc[s], max(1, occ_blksize)) for s in [0,1]]

    log.debug('occ blksize for %s loop: %d/%d %d/%d', mp.__class__.__name__,
              occ_blksize[0], nocc[0], occ_blksize[1], nocc[1])

    cput1 = (logger.process_clock(), logger.perf_counter())

    emp2_ss = emp2_os = 0
    # same spin
    drv = libmp.MP2_contract_d
    for s in [0,1]:
        s_t2 = 0 if s == 0 else 2
        moevv = lib.asarray(vir_energy[s][:,None] + vir_energy[s], order='C')
        for ibatch,(i0,i1) in enumerate(lib.prange(0,nocc[s],occ_blksize[s])):
            nocci = i1-i0
            iaL = eris.get_occ_blk(s,i0,i1)
            for jbatch,(j0,j1) in enumerate(lib.prange(0,nocc[s],occ_blksize[s])):
                noccj = j1-j0
                if ibatch == jbatch:
                    jbL = iaL
                else:
                    jbL = eris.get_occ_blk(s,j0,j1)

                ed = np.zeros(1, dtype=np.float64)
                ex = np.zeros(1, dtype=np.float64)
                moeoo_block = np.asarray(
                    occ_energy[s][i0:i1,None] + occ_energy[s][j0:j1], order='C')
                s2symm = 1
                t2_ex = True
                drv(
                    ed.ctypes.data_as(ctypes.c_void_p),
                    ex.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(s2symm),
                    iaL.ctypes.data_as(ctypes.c_void_p),
                    jbL.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(i0), ctypes.c_int(j0),
                    ctypes.c_int(nocci), ctypes.c_int(noccj),
                    ctypes.c_int(nocc[s]), ctypes.c_int(nvir[s]), ctypes.c_int(naux),
                    moeoo_block.ctypes.data_as(ctypes.c_void_p),
                    moevv.ctypes.data_as(ctypes.c_void_p),
                    t2_ptr[s_t2], ctypes.c_int(t2_ex)
                )
                emp2_ss += (ed[0] + ex[0]) * 0.5

                jbL = None
            iaL = None

            cput1 = log.timer_debug1('(sa,sb) = (%d,%d)  i-block [%d:%d]/%d' % (s,s,i0,i1,nocc[s]),
                                     *cput1)

    # opposite spin
    sa, sb = 0, 1
    drv = libmp.MP2_OS_contract_d
    moevv = lib.asarray(vir_energy[sa][:,None] + vir_energy[sb], order='C')
    for ibatch,(i0,i1) in enumerate(lib.prange(0,nocc[sa],occ_blksize[sa])):
        nocci = i1-i0
        iaL = eris.get_occ_blk(sa,i0,i1)
        for jbatch,(j0,j1) in enumerate(lib.prange(0,nocc[sb],occ_blksize[sb])):
            noccj = j1-j0
            jbL = eris.get_occ_blk(sb,j0,j1)

            ed = np.zeros(1, dtype=np.float64)
            moeoo_block = np.asarray(
                occ_energy[sa][i0:i1,None] + occ_energy[sb][j0:j1], order='C')
            drv(
                ed.ctypes.data_as(ctypes.c_void_p),
                iaL.ctypes.data_as(ctypes.c_void_p),
                jbL.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(i0), ctypes.c_int(j0),
                ctypes.c_int(nocci), ctypes.c_int(noccj),
                ctypes.c_int(nocc[sa]), ctypes.c_int(nocc[sb]),
                ctypes.c_int(nvir[sa]), ctypes.c_int(nvir[sb]),
                ctypes.c_int(naux),
                moeoo_block.ctypes.data_as(ctypes.c_void_p),
                moevv.ctypes.data_as(ctypes.c_void_p),
                t2_ptr[1]
            )
            emp2_os += ed[0]

            jbL = None
        iaL = None

        cput1 = log.timer_debug1('(sa,sb) = (%d,%d)  i-block [%d:%d]/%d' % (sa,sb,i0,i1,nocc[sa]),
                                 *cput1)

    emp2_ss = emp2_ss.real
    emp2_os = emp2_os.real
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)

    return emp2, t2


class DFUMP2(ump2.UMP2):
    _keys = dfmp2.DFRMP2._keys

    __init__ = dfmp2.DFRMP2.__init__

    get_nocc = ump2.get_nocc
    get_nmo = ump2.get_nmo
    get_frozen_mask = ump2.get_frozen_mask

    kernel = ump2.UMP2.kernel

    make_fno = ump2.make_fno
    make_rdm1 = make_rdm1
    make_rdm2 = make_rdm2

    reset = dfmp2.DFRMP2.reset

    def split_mo_coeff(self, mo_coeff=None):
        if mo_coeff is None: mo_coeff = self.mo_coeff
        masks = _mo_splitter(self)
        return [[mo_coeff[s][:,m] for m in masks[s]] for s in [0,1]]

    def split_mo_energy(self, mo_energy=None):
        if mo_energy is None: mo_energy = self.mo_energy
        masks = _mo_splitter(self)
        return [[mo_energy[s][m] for m in masks[s]] for s in [0,1]]

    def split_mo_occ(self, mo_occ=None):
        if mo_occ is None: mo_occ = self.mo_occ
        masks = _mo_splitter(self)
        return [[mo_occ[s][m] for m in masks[s]] for s in [0,1]]

    def ao2mo(self, mo_coeff=None, ovL=None, ovL_to_save=None):
        return _make_df_eris(self, mo_coeff, ovL, ovL_to_save)

    def nuc_grad_method(self):
        raise NotImplementedError

    # For non-canonical MP2
    def update_amps(self, t2, eris):
        raise NotImplementedError

    def init_amps(self, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2):
        return kernel(self, mo_energy, mo_coeff, eris, with_t2)

    Gradients = NotImplemented

MP2 = UMP2 = DFUMP2

from pyscf import scf
scf.uhf.UHF.DFMP2 = lib.class_as_method(DFUMP2)
del (WITH_T2)


def _make_df_eris(mp, mo_coeff=None, ovL=None, ovL_to_save=None, verbose=None):
    log = logger.new_logger(mp, verbose)

    with_df = getattr(mp, 'with_df', None)
    assert( with_df is not None )

    if with_df._cderi is None:
        log.debug('Caching ovL-type integrals directly')
        if with_df.auxmol is None:
            with_df.auxmol = df.addons.make_auxmol(with_df.mol, with_df.auxbasis)
    else:
        log.debug('Caching ovL-type integrals by transforming saved AO 3c integrals.')

    if mo_coeff is None: mo_coeff = mp.mo_coeff
    split_mo_coeff = mp.split_mo_coeff()
    occ_coeff = [x[1] for x in split_mo_coeff]
    vir_coeff = [x[2] for x in split_mo_coeff]

    # determine incore or outcore
    nocc = np.asarray([x.shape[1] for x in occ_coeff])
    nvir = np.asarray([x.shape[1] for x in vir_coeff])
    naux = with_df.get_naoaux()

    if ovL is not None:
        if isinstance(ovL, (np.ndarray,list,tuple)):
            outcore = False
        elif isinstance(ovL, str):
            outcore = True
        else:
            log.error('Unknown data type %s for input `ovL` (should be np.ndarray or str).',
                      type(ovL))
            raise TypeError
    else:
        mem_now = mp.max_memory - lib.current_memory()[0]
        mem_df = sum(nocc*nvir)*naux*8/1024**2.
        log.debug('ao2mo est mem= %.2f MB  avail mem= %.2f MB', mem_df, mem_now)
        # DEBUG:
        if mp.force_outcore:
            outcore = True
        else:
            outcore = (ovL_to_save is not None) or (mem_now*0.8 < mem_df)
    log.debug('ovL-type integrals are cached %s', 'outcore' if outcore else 'incore')

    if outcore:
        eris = _DFOUTCOREERIS(with_df, occ_coeff, vir_coeff, mp.max_memory,
                              ovL=ovL, ovL_to_save=ovL_to_save,
                              verbose=log.verbose, stdout=log.stdout)
    else:
        eris = _DFINCOREERIS(with_df, occ_coeff, vir_coeff, mp.max_memory,
                             ovL=ovL,
                             verbose=log.verbose, stdout=log.stdout)
    eris.build()

    return eris


class _DFINCOREERIS:
    def __init__(self, with_df, occ_coeff, vir_coeff, max_memory, ovL=None,
                 verbose=None, stdout=None):
        self.with_df = with_df
        self.occ_coeff = occ_coeff
        self.vir_coeff = vir_coeff

        self.max_memory = max_memory
        self.verbose = verbose
        self.stdout = stdout

        self.dtype = np.result_type(*self.occ_coeff)
        assert( self.dtype == np.float64 )  # FIXME: support complex
        self.dsize = 8

        self.ovL = ovL

    @property
    def nocc(self):
        return [x.shape[1] for x in self.occ_coeff]
    @property
    def nvir(self):
        return [x.shape[1] for x in self.vir_coeff]
    @property
    def naux(self):
        return self.ovL[0].shape[-1]

    def build(self):
        log = logger.new_logger(self)
        if self.ovL is None:
            if self.with_df._cderi is None:
                self.ovL = _init_mp_df_eris_direct(self.with_df, self.occ_coeff, self.vir_coeff,
                                                   self.max_memory, log=log)
            else:
                self.ovL = _init_mp_df_eris(self.with_df, self.occ_coeff, self.vir_coeff,
                                            self.max_memory, log=log)

    def get_occ_blk(self, s,i0,i1):
        nvir, naux = self.nvir[s], self.naux
        return np.asarray(self.ovL[s][i0*nvir:i1*nvir], order='C').reshape(i1-i0,nvir,naux)
    def get_ov_blk(self, s,ia0,ia1):
        return np.asarray(self.ovL[s][ia0:ia1], order='C')


class _DFOUTCOREERIS(_DFINCOREERIS):
    def __init__(self, with_df, occ_coeff, vir_coeff, max_memory, ovL=None, ovL_to_save=None,
                 verbose=None, stdout=None):
        _DFINCOREERIS.__init__(self, with_df, occ_coeff, vir_coeff, max_memory, None,
                               verbose, stdout)

        self._ovL = ovL
        self._ovL_to_save = ovL_to_save

    def build(self):
        log = logger.new_logger(self)
        with_df = self.with_df
        if self._ovL is None:
            if isinstance(self._ovL_to_save, str):
                self.feri = lib.H5FileWrap(self._ovL_to_save, 'w')
            else:
                self.feri = lib.H5TmpFile()
            log.debug('ovL is saved to %s', self.feri.filename)
            if with_df._cderi is None:
                _init_mp_df_eris_direct(with_df, self.occ_coeff, self.vir_coeff, self.max_memory,
                                        h5obj=self.feri, log=log)
            else:
                _init_mp_df_eris(with_df, self.occ_coeff, self.vir_coeff, self.max_memory,
                                 h5obj=self.feri, log=log)
            self.ovL = [self.feri[f'ovL{s}'] for s in [0,1]]
        elif isinstance(self._ovL, str):
            self.feri = h5py.File(self._ovL, 'r')
            log.debug('ovL is read from %s', self.feri.filename)
            assert( 'ovL0' in self.feri and 'ovL1' in self.feri )
            self.ovL = [self.feri[f'ovL{s}'] for s in [0,1]]
        else:
            raise RuntimeError

def _init_mp_df_eris(with_df, occ_coeff, vir_coeff, max_memory, h5obj=None, log=None):
    if log is None: log = logger.new_logger(with_df)

    nao = occ_coeff[0].shape[0]
    nocc = np.asarray([x.shape[1] for x in occ_coeff])
    nvir = np.asarray([x.shape[1] for x in vir_coeff])
    noccmax = max(nocc)
    nvirmax = max(nvir)
    nmo = [nocc[s] + nvir[s] for s in [0,1]]
    nao_pair = nao**2
    naux = with_df.get_naoaux()

    dtype = np.result_type(*occ_coeff)
    assert( dtype == np.float64 )
    dsize = 8

    mo = [np.asarray(np.hstack((occ_coeff[s],vir_coeff[s])), order='F') for s in [0,1]]

    mem_avail = max_memory - lib.current_memory()[0]

    if h5obj is None:   # incore
        ovL = [np.empty((nocc[s]*nvir[s],naux), dtype=dtype) for s in [0,1]]
        mem_avail -= sum(nocc*nvir)*naux * dsize/1e6
    else:
        ovL = []
        for s in [0,1]:
            ovL_shape = (nocc[s]*nvir[s],naux)
            ovL.append(h5obj.create_dataset(f'ovL{s}', ovL_shape, dtype=dtype,
                                            chunks=(1,*ovL_shape[1:])))

    if isinstance(ovL, np.ndarray):
        # incore: batching aux (OV + Nao_pair) * [X] = M
        mem_auxblk = (nao_pair+noccmax*nvirmax) * dsize/1e6
        aux_blksize = min(naux, max(1, int(np.floor(mem_avail*0.5 / mem_auxblk))))
        log.debug('aux blksize for incore ao2mo: %d/%d', aux_blksize, naux)
        buf = np.empty(aux_blksize*noccmax*nvirmax, dtype=dtype)
        ijslice = [(0, nocc[s], nocc[s], nmo[s]) for s in [0,1]]

        p1 = 0
        for Lpq in with_df.loop(blksize=aux_blksize):
            p0, p1 = p1, p1+Lpq.shape[0]
            for s in [0,1]:
                out = _ao2mo.nr_e2(Lpq, mo[s], ijslice[s], aosym='s2', out=buf)
                ovL[s][:,p0:p1] = out.T
                out = None
            Lpq = None
        buf = None
    else:
        # outcore: batching occ [O]XV and aux ([O]V + Nao_pair)*[X]
        mem_occblk = naux*nvirmax * dsize/1e6
        occ_blksize = min(noccmax, max(1, int(np.floor(mem_avail*0.6 / mem_occblk))))
        mem_auxblk = (occ_blksize*nvirmax+nao_pair) * dsize/1e6
        aux_blksize = min(naux, max(1, int(np.floor(mem_avail*0.3 / mem_auxblk))))
        log.debug('occ blksize for outcore ao2mo: %d/%d', occ_blksize, noccmax)
        log.debug('aux blksize for outcore ao2mo: %d/%d', aux_blksize, naux)
        buf = np.empty(naux*occ_blksize*nvirmax, dtype=dtype)
        buf2 = np.empty(aux_blksize*occ_blksize*nvirmax, dtype=dtype)

        for s in [0,1]:
            for i0,i1 in lib.prange(0,nocc[s],occ_blksize):
                nocci = i1-i0
                ijslice = (i0,i1,nocc[s],nmo[s])
                p1 = 0
                OvL = np.ndarray((nocci*nvir[s],naux), dtype=dtype, buffer=buf)
                for Lpq in with_df.loop(blksize=aux_blksize):
                    p0, p1 = p1, p1+Lpq.shape[0]
                    out = _ao2mo.nr_e2(Lpq, mo[s], ijslice, aosym='s2', out=buf2)
                    OvL[:,p0:p1] = out.T
                    Lpq = out = None
                # this avoids slow operations like ovL[i0:i1,:,p0:p1] = ...
                ovL[s][i0*nvir[s]:i1*nvir[s]] = OvL
                OvL = None
        buf = buf2 = None

    return ovL


def _init_mp_df_eris_direct(with_df, occ_coeff, vir_coeff, max_memory, h5obj=None, log=None):
    from pyscf import gto
    from pyscf.df.incore import fill_2c2e
    from pyscf.ao2mo.outcore import balance_partition

    if log is None: log = logger.new_logger(with_df)

    mol = with_df.mol
    auxmol = with_df.auxmol
    nbas = mol.nbas
    nao = occ_coeff[0].shape[0]
    nocc = np.asarray([x.shape[1] for x in occ_coeff])
    nvir = np.asarray([x.shape[1] for x in vir_coeff])
    noccmax = max(nocc)
    nvirmax = max(nvir)
    nmo = [nocc[s] + nvir[s] for s in [0,1]]
    nao_pair = nao*(nao+1)//2
    naoaux = auxmol.nao_nr()

    dtype = np.result_type(*occ_coeff)
    assert( dtype == np.float64 )
    dsize = 8

    mo = [np.asarray(np.hstack((occ_coeff[s],vir_coeff[s])), order='F') for s in [0,1]]
    ijslice = [(0, nocc[s], nocc[s], nmo[s]) for s in [0,1]]

    tspans = np.zeros((5,2))
    tnames = ['j2c', 'j3c', 'xform', 'save', 'fit']
    tick = (logger.process_clock(), logger.perf_counter())
    # precompute for fitting
    j2c = fill_2c2e(mol, auxmol)
    try:
        m2c = scipy.linalg.cholesky(j2c, lower=True)
        tag = 'cd'
    except scipy.linalg.LinAlgError:
        e, u = np.linalg.eigh(j2c)
        cond = abs(e).max() / abs(e).min()
        keep = abs(e) > THRESH_LINDEP
        log.debug('cond(j2c) = %g', cond)
        log.debug('keep %d/%d cderi vectors', np.count_nonzero(keep), keep.size)
        e = e[keep]
        u = u[:,keep]
        m2c = lib.dot(u*e**-0.5, u.T.conj())
        tag = 'eig'
    j2c = None
    naux = m2c.shape[1]
    tock = (logger.process_clock(), logger.perf_counter())
    tspans[0] += np.asarray(tock) - np.asarray(tick)

    mem_avail = max_memory - lib.current_memory()[0]

    incore = h5obj is None
    if incore:
        ovL = [np.empty((nocc[s]*nvir[s],naux), dtype=dtype) for s in [0,1]]
        mem_avail -= sum([x.size for x in ovL]) * dsize / 1e6
    else:
        ovL = []
        for s in [0,1]:
            ovL_shape = (nocc[s]*nvir[s],naux)
            ovL.append( h5obj.create_dataset(f'ovL{s}', ovL_shape, dtype=dtype,
                                             chunks=(1,*ovL_shape[1:])) )
        h5tmp = lib.H5TmpFile()
        Lov0 = []
        for s in [0,1]:
            Lov0_shape = (naoaux,nocc[s]*nvir[s])
            Lov0.append( h5tmp.create_dataset(f'Lov0{s}', Lov0_shape, dtype=dtype,
                                              chunks=(1,*Lov0_shape[1:])) )

    # buffer
    mem_blk = nao_pair*2 * dsize / 1e6
    aux_blksize = max(1, min(naoaux, int(np.floor(mem_avail*0.7 / mem_blk))))
    auxshl_range = balance_partition(auxmol.ao_loc, aux_blksize)
    auxlen = max([x[2] for x in auxshl_range])
    log.info('mem_avail = %.2f  mem_blk = %.2f  auxlen = %d', mem_avail, mem_blk, auxlen)
    buf0 = np.empty(auxlen*nao_pair, dtype=dtype)
    buf0T = np.empty(auxlen*nao_pair, dtype=dtype)

    # precompute for j3c
    comp = 1
    aosym = 's2ij'
    int3c = gto.moleintor.ascint3(mol._add_suffix('int3c2e'))
    atm_f, bas_f, env_f = gto.mole.conc_env(mol._atm, mol._bas, mol._env,
                                            auxmol._atm, auxmol._bas, auxmol._env)
    ao_loc_f = gto.moleintor.make_loc(bas_f, int3c)
    cintopt = gto.moleintor.make_cintopt(atm_f, bas_f, env_f, int3c)

    def calc_j3c_ao(kshl0, kshl1):
        shls_slice = (0, nbas, 0, nbas, nbas+kshl0, nbas+kshl1)
        pqL = gto.moleintor.getints3c(int3c, atm_f, bas_f, env_f, shls_slice, comp,
                                      aosym, ao_loc_f, cintopt, out=buf0)
        Lpq = lib.transpose(pqL, out=buf0T)
        pqL = None
        return Lpq

    # transform
    k1 = 0
    for auxshl_rg in auxshl_range:
        kshl0, kshl1, dk = auxshl_rg
        k0, k1 = k1, k1+dk
        log.debug('kshl = [%d:%d/%d]  [%d:%d/%d]', kshl0, kshl1, auxmol.nbas, k0, k1, naoaux)
        tick = (logger.process_clock(), logger.perf_counter())
        lpq = calc_j3c_ao(kshl0, kshl1)
        tock = (logger.process_clock(), logger.perf_counter())
        tspans[1] += np.asarray(tock) - np.asarray(tick)
        for s in [0,1]:
            lov = _ao2mo.nr_e2(lpq, mo[s], ijslice[s], aosym='s2', out=buf0)
            tick = (logger.process_clock(), logger.perf_counter())
            tspans[2] += np.asarray(tick) - np.asarray(tock)
            if incore:
                ovl = lib.transpose(lov, out=buf0T)
                ovL[s][:,k0:k1] = ovl
                ovl = None
            else:
                Lov0[s][k0:k1] = lov
            lov = None
            tock = (logger.process_clock(), logger.perf_counter())
            tspans[3] += np.asarray(tock) - np.asarray(tick)
        lpq = None
    buf0 = buf0T = None

    tick = (logger.process_clock(), logger.perf_counter())
    # fit
    if tag == 'cd': drv = getattr(libmp, 'trisolve_parallel_grp', None)
    if incore:
        if tag == 'cd':
            if drv is None:
                for s in [0,1]:
                    scipy.linalg.solve_triangular(m2c, ovL[s].T, lower=True,
                                                  overwrite_b=True, check_finite=False).T
            else:
                assert m2c.flags.f_contiguous
                grpfac = 10
                for s in [0,1]:
                    drv(
                        m2c.ctypes.data_as(ctypes.c_void_p),
                        ovL[s].ctypes.data_as(ctypes.c_void_p),
                        ctypes.c_int(naux),
                        ctypes.c_int(nocc[s]*nvir[s]),
                        ctypes.c_int(grpfac)
                    )
        else:
            mem_blk = nvirmax*naux * dsize/1e6
            occ_blksize0 = max(1, min(noccmax, int(np.floor(mem_avail*0.5/mem_blk))))
            buf = np.empty(occ_blksize0*nvirmax*naux, dtype=dtype)
            for s in [0,1]:
                occ_blksize = min(nocc[s], occ_blksize0)
                ovL[s] = ovL[s].reshape(-1)
                nvxao = nvir[s]*naoaux
                nvx = nvir[s]*naux
                for i0,i1 in lib.prange(0,nocc[s],occ_blksize):
                    nocci = i1-i0
                    out = np.ndarray((nocci*nvir[s],naux), dtype=dtype, buffer=buf)
                    lib.dot(ovL[s][i0*nvxao:i1*nvxao].reshape(nocci*nvir[s],naoaux), m2c, c=out)
                    ovL[s][i0*nvx:i1*nvx] = out.reshape(-1)
                ovL[s] = ovL[s][:nocc[s]*nvx].reshape(nocc[s]*nvir[s],naux)
            buf = None
    else:
        mem_blk = nvirmax*naoaux * dsize / 1e6
        occ_blksize0 = max(1, min(noccmax, int(np.floor(mem_avail*0.4/mem_blk))))
        for s in [0,1]:
            occ_blksize = min(nocc[s], occ_blksize0)
            nvxao = nvir[s]*naoaux
            nvx = nvir[s]*naux
            for i0,i1 in lib.prange(0, nocc[s], occ_blksize):
                nocci = i1-i0
                ivL = np.asarray(Lov0[s][:,i0*nvir[s]:i1*nvir[s]].T, order='C')
                if tag == 'cd':
                    if drv is None:
                        ivL = scipy.linalg.solve_triangular(m2c, ivL.T, lower=True,
                                                            overwrite_b=True, check_finite=False).T
                    else:
                        assert m2c.flags.f_contiguous
                        grpfac = 10
                        drv(
                            m2c.ctypes.data_as(ctypes.c_void_p),
                            ivL.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(naux),
                            ctypes.c_int(nocci*nvir[s]),
                            ctypes.c_int(grpfac)
                        )
                else:
                    ivL = lib.dot(ivL.reshape(nocci*nvir[s],naoaux), m2c)
                ovL[s][i0*nvir[s]:i1*nvir[s]] = ivL

        for s in [0,1]:
            del h5tmp[f'Lov0{s}']
        h5tmp.close()
        Lov0 = None
    tock = (logger.process_clock(), logger.perf_counter())
    tspans[4] += np.asarray(tock) - np.asarray(tick)

    for tspan,tname in zip(tspans,tnames):
        log.debug('ao2mo CPU time for %-10s  %9.2f sec  wall time %9.2f sec', tname, *tspan)
    log.info('')

    return ovL



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py

def get_version():
    topdir = os.path.abspath(os.path.join(__file__, '..'))
    with open(os.path.join(topdir, 'pyscf', '__init__.py'), 'r') as f:
        for line in f.readlines():
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise ValueError("Version string not found")
VERSION = get_version()


def get_platform():
    from distutils.util import get_platform
    platform = get_platform()
    if sys.platform == 'darwin':
        arch = os.getenv('CMAKE_OSX_ARCHITECTURES')
        if arch:
            osname = platform.rsplit('-', 1)[0]
            if ';' in arch:
                platform = f'{osname}-universal2'
            else:
                platform = f'{osname}-{arch}'
        elif os.getenv('_PYTHON_HOST_PLATFORM'):
            # the cibuildwheel environment
            platform = os.getenv('_PYTHON_HOST_PLATFORM')
            if platform.endswith('arm64'):
                os.putenv('CMAKE_OSX_ARCHITECTURES', 'arm64')
            elif platform.endswith('x86_64'):
                os.putenv('CMAKE_OSX_ARCHITECTURES', 'x86_64')
            else:
                os.putenv('CMAKE_OSX_ARCHITECTURES', 'arm64;x86_64')
    return platform

class CMakeBuildPy(build_py):
    def run(self):
        self.plat_name = get_platform()
        self.build_base = 'build'
        self.build_lib = os.path.join(self.build_base, 'lib')
        self.build_temp = os.path.join(self.build_base, f'temp.{self.plat_name}')

        self.announce('Configuring extensions', level=3)
        src_dir = os.path.abspath(os.path.join(__file__, '..', 'pyscf', 'lib'))
        cmd = ['cmake', f'-S{src_dir}', f'-B{self.build_temp}']
        configure_args = os.getenv('CMAKE_CONFIGURE_ARGS')
        if configure_args:
            cmd.extend(configure_args.split(' '))
        self.spawn(cmd)

        self.announce('Building binaries', level=3)
        # By default do not use high level parallel compilation.
        # OOM may be triggered when compiling certain functionals in libxc.
        # Set the shell variable CMAKE_BUILD_PARALLEL_LEVEL=n to enable
        # parallel compilation.
        cmd = ['cmake', '--build', self.build_temp]
        build_args = os.getenv('CMAKE_BUILD_ARGS')
        if build_args:
            cmd.extend(build_args.split(' '))
        if self.dry_run:
            self.announce(' '.join(cmd))
        else:
            self.spawn(cmd)
        super().run()

# build_py will produce plat_name = 'any'. Patch the bdist_wheel to change the
# platform tag because the C extensions are platform dependent.
# For setuptools<70
from wheel.bdist_wheel import bdist_wheel
initialize_options_1 = bdist_wheel.initialize_options
def initialize_with_default_plat_name(self):
    initialize_options_1(self)
    self.plat_name = get_platform()
    self.plat_name_supplied = True
bdist_wheel.initialize_options = initialize_with_default_plat_name

# For setuptools>=70
try:
    from setuptools.command.bdist_wheel import bdist_wheel
    initialize_options_2 = bdist_wheel.initialize_options
    def initialize_with_default_plat_name(self):
        initialize_options_2(self)
        self.plat_name = get_platform()
        self.plat_name_supplied = True
    bdist_wheel.initialize_options = initialize_with_default_plat_name
except ImportError:
    pass

# scipy bugs
# https://github.com/scipy/scipy/issues/12533
_scipy_version = 'scipy!=1.5.0,!=1.5.1'
if sys.platform == 'darwin':
    if sys.version_info < (3, 8):
        _scipy_version = 'scipy<=1.1.0'
    else:
        print('scipy>1.1.0 may crash when calling scipy.linalg.eigh. '
              '(Issues https://github.com/scipy/scipy/issues/15362 '
              'https://github.com/scipy/scipy/issues/16151)')

setup(
    version=VERSION,
    #package_dir={'pyscf': 'pyscf'},  # packages are under directory pyscf
    #include *.so *.dat files. They are now placed in MANIFEST.in
    #package_data={'': ['*.so', '*.dylib', '*.dll', '*.dat']},
    include_package_data=True,  # include everything in source control
    packages=find_packages(exclude=['*test*', '*examples*']),
    cmdclass={'build_py': CMakeBuildPy},
)
